# stdlib

# relative
from ...serde.serializable import serializable
from ...store.document_store import DocumentStore
from ...store.linked_obj import LinkedObject
from ...types.uid import UID
from ...util.telemetry import instrument
from ..context import AuthedServiceContext
from ..notification.notification_service import NotificationService
from ..notification.notifications import CreateNotification
from ..response import SyftError
from ..response import SyftNotReady
from ..response import SyftSuccess
from ..service import AbstractService
from ..service import SERVICE_TO_TYPES
from ..service import TYPE_TO_SERVICE
from ..service import service_method
from ..user.user_roles import GUEST_ROLE_LEVEL
from ..user.user_roles import ONLY_DATA_SCIENTIST_ROLE_LEVEL
from ..user.user_roles import ServiceRole
from .project import Project
from .project import ProjectEvent
from .project import ProjectRequest
from .project import ProjectSubmit
from .project import create_project_hash
from .project_stash import ProjectStash


@instrument
@serializable(canonical_name="ProjectService", version=1)
class ProjectService(AbstractService):
    store: DocumentStore
    stash: ProjectStash

    def __init__(self, store: DocumentStore) -> None:
        self.store = store
        self.stash = ProjectStash(store=store)

    @service_method(
        path="project.can_create_project",
        name="can_create_project",
        roles=ONLY_DATA_SCIENTIST_ROLE_LEVEL,
    )
    def can_create_project(self, context: AuthedServiceContext) -> bool | SyftError:
        user_service = context.server.get_service("userservice")
        role = user_service.get_role_for_credentials(credentials=context.credentials)
        if role == ServiceRole.DATA_SCIENTIST:
            return True
        return SyftError(message="User cannot create projects")

    @service_method(
        path="project.create_project",
        name="create_project",
        roles=ONLY_DATA_SCIENTIST_ROLE_LEVEL,
    )
    def create_project(
        self, context: AuthedServiceContext, project: ProjectSubmit
    ) -> SyftSuccess | SyftError:
        """Start a Project"""

        check_role = self.can_create_project(context)
        if isinstance(check_role, SyftError):
            return check_role

        try:
            # Check if the project with given id already exists
            project_id_check = self.stash.get_by_uid(
                credentials=context.server.verify_key, uid=project.id
            )

            if project_id_check.is_err():
                return SyftError(message=f"{project_id_check.err()}")

            if project_id_check.ok() is not None:
                return SyftError(
                    message=f"Project with id: {project.id} already exists."
                )

            project_obj: Project = project.to(Project, context=context)

            # Updating the leader server route of the project object
            # In case the current server, is the leader, they would input their server route
            # For the followers, they would check if the leader is their server peer
            # using the leader's verify_key
            # If the follower do not have the leader as its peer in its routes
            # They would raise as error
            leader_server = project_obj.state_sync_leader

            # If the current server is a follower
            # For followers the leader server route is retrieved from its peer
            if leader_server.verify_key != context.server.verify_key:
                network_service = context.server.get_service("networkservice")
                peer = network_service.stash.get_by_verify_key(
                    credentials=context.server.verify_key,
                    verify_key=leader_server.verify_key,
                )
                if peer.is_err():
                    this_server_id = (
                        context.server.id.short() if context.server.id else ""
                    )
                    return SyftError(
                        message=(
                            f"Leader Server(id={leader_server.id.short()}) is not a "
                            f"peer of this Server(id={this_server_id})"
                        )
                    )
                leader_server_peer = peer.ok()
            else:
                # for the leader server, as it does not have route information to itself
                # we rely on the data scientist to provide the route
                # the route is then validated by the leader
                if project.leader_server_route is not None:
                    leader_server_peer = (
                        project.leader_server_route.validate_with_context(
                            context=context
                        )
                    )
                else:
                    return SyftError(
                        message=f"project {project}'s leader_server_route is None"
                    )

            project_obj.leader_server_peer = leader_server_peer

            # This should always be the last call before flushing to DB
            project_obj.start_hash = create_project_hash(project_obj)[1]

            result = self.stash.set(context.credentials, project_obj)
            if result.is_err():
                return SyftError(message=str(result.err()))

            project_obj_store = result.ok()
            project_obj_store = self.add_signing_key_to_project(
                context, project_obj_store
            )

            return project_obj_store

        except Exception as e:
            print("Failed to submit Project", e)
            raise e

    @service_method(
        path="project.add_event",
        name="add_event",
        roles=GUEST_ROLE_LEVEL,
    )
    def add_event(
        self, context: AuthedServiceContext, project_event: ProjectEvent
    ) -> SyftSuccess | SyftError:
        """To add events to a projects"""

        # Event object should be received from the leader of the project

        # retrieve the project object by server verify key
        project_obj = self.stash.get_by_uid(
            context.server.verify_key, uid=project_event.project_id
        )
        if project_obj.is_err():
            return SyftError(message=str(project_obj.err()))

        project: Project = project_obj.ok()
        if project.state_sync_leader.verify_key == context.server.verify_key:
            return SyftError(
                message="Project Events should be passed to leader by broadcast endpoint"
            )
        if context.credentials != project.state_sync_leader.verify_key:
            return SyftError(message="Only the leader of the project can add events")

        project.events.append(project_event)
        project.event_id_hashmap[project_event.id] = project_event

        message_result = self.check_for_project_request(project, project_event, context)
        if isinstance(message_result, SyftError):
            return message_result

        # updating the project object using root verify key of server
        result = self.stash.update(context.server.verify_key, project)

        if result.is_err():
            return SyftError(message=str(result.err()))
        return SyftSuccess(
            message=f"Project event {project_event.id} added successfully "
        )

    @service_method(
        path="project.broadcast_event",
        name="broadcast_event",
        roles=GUEST_ROLE_LEVEL,
    )
    def broadcast_event(
        self, context: AuthedServiceContext, project_event: ProjectEvent
    ) -> SyftSuccess | SyftError:
        """To add events to a projects"""
        # Only the leader of the project could add events to the projects
        # Any Event to be added to the project should be sent to the leader of the project
        # The leader broadcasts the event to all the members of the project

        project_obj = self.stash.get_by_uid(
            context.server.verify_key, uid=project_event.project_id
        )

        if project_obj.is_err():
            return SyftError(message=str(project_obj.err()))

        project = project_obj.ok()
        if not project.has_permission(context.credentials):
            return SyftError(message="User does not have permission to add events")

        if project.state_sync_leader.verify_key != context.server.verify_key:
            return SyftError(
                message="Only the leader of the project can broadcast events"
            )

        if project_event.seq_no is None:
            return SyftError(message=f"{project_event}.seq_no is None")
        if project_event.seq_no <= len(project.events) and len(project.events) > 0:
            return SyftNotReady(message="Project out of sync event")
        if project_event.seq_no > len(project.events) + 1:
            return SyftError(message="Project event out of order!")

        project.events.append(project_event)
        project.event_id_hashmap[project_event.id] = project_event

        message_result = self.check_for_project_request(project, project_event, context)
        if isinstance(message_result, SyftError):
            return message_result

        # Broadcast the event to all the members of the project
        network_service = context.server.get_service("networkservice")
        for member in project.members:
            if member.verify_key != context.server.verify_key:
                # Retrieving the ServerPeer Object to communicate with the server
                peer = network_service.stash.get_by_verify_key(
                    credentials=context.server.verify_key,
                    verify_key=member.verify_key,
                )

                if peer.is_err():
                    return SyftError(
                        message=f"Leader server does not have peer {member.name}-{member.id.short()}"
                        + " Kindly exchange routes with the peer"
                    )
                peer = peer.ok()
                remote_client = peer.client_with_context(context=context)
                if remote_client.is_err():
                    return SyftError(
                        message=f"Failed to create remote client for peer: "
                        f"{peer.id}. Error: {remote_client.err()}"
                    )
                remote_client = remote_client.ok()

                event_result = remote_client.api.services.project.add_event(
                    project_event
                )
                if isinstance(event_result, SyftError):
                    return event_result

        result = self.stash.update(context.server.verify_key, project)

        if result.is_err():
            return SyftError(message=str(result.err()))
        return SyftSuccess(message="Successfully Broadcasted Event")

    @service_method(
        path="project.sync",
        name="sync",
        roles=GUEST_ROLE_LEVEL,
    )
    def sync(
        self, context: AuthedServiceContext, project_id: UID, seq_no: int
    ) -> list[ProjectEvent] | SyftError:
        """To fetch unsynced events from the project"""

        # Event object should be received from the leader of the project

        # retrieve the project object by server verify key
        project_obj = self.stash.get_by_uid(context.server.verify_key, uid=project_id)
        if project_obj.is_err():
            return SyftError(message=str(project_obj.err()))

        project: Project = project_obj.ok()
        if project.state_sync_leader.verify_key != context.server.verify_key:
            return SyftError(
                message="Project Events should be synced only with the leader"
            )

        if not project.has_permission(context.credentials):
            return SyftError(message="User does not have permission to sync events")

        if seq_no < 0:
            return SyftError(message="Input seq_no should be a non negative integer")

        # retrieving unsycned events based on seq_no
        return project.events[seq_no:]

    @service_method(path="project.get_all", name="get_all", roles=GUEST_ROLE_LEVEL)
    def get_all(self, context: AuthedServiceContext) -> list[Project] | SyftError:
        result = self.stash.get_all(
            context.credentials,
        )
        if result.is_err():
            return SyftError(message=str(result.err()))

        projects = result.ok()

        for idx, project in enumerate(projects):
            result = self.add_signing_key_to_project(context, project)
            if isinstance(result, SyftError):
                return result
            projects[idx] = result

        return projects

    @service_method(
        path="project.get_by_name",
        name="get_by_name",
        roles=GUEST_ROLE_LEVEL,
    )
    def get_by_name(
        self, context: AuthedServiceContext, name: str
    ) -> Project | SyftError:
        result = self.stash.get_by_name(context.credentials, project_name=name)
        if result.is_err():
            return SyftError(message=str(result.err()))
        elif result.ok():
            project = result.ok()
            return self.add_signing_key_to_project(context, project)
        return SyftError(message=f'Project(name="{name}") does not exist')

    @service_method(
        path="project.get_by_uid",
        name="get_by_uid",
        roles=GUEST_ROLE_LEVEL,
    )
    def get_by_uid(
        self, context: AuthedServiceContext, uid: UID
    ) -> Project | SyftError:
        result = self.stash.get_by_uid(
            credentials=context.server.verify_key,
            uid=uid,
        )
        if result.is_err():
            return SyftError(message=str(result.err()))
        elif result.ok():
            return result.ok()
        return SyftError(message=f'Project(id="{uid}") does not exist')

    def add_signing_key_to_project(
        self, context: AuthedServiceContext, project: Project
    ) -> Project | SyftError:
        # Automatically infuse signing key of user
        # requesting get_all() or creating the project object

        user_service = context.server.get_service("userservice")
        user = user_service.stash.get_by_verify_key(
            credentials=context.credentials, verify_key=context.credentials
        )
        if user.is_err():
            return SyftError(message=str(user.err()))

        user = user.ok()
        if not user:
            return SyftError(message="User not found! Kindly register user first")

        project.user_signing_key = user.signing_key

        return project

    def check_for_project_request(
        self,
        project: Project,
        project_event: ProjectEvent,
        context: AuthedServiceContext,
    ) -> SyftSuccess | SyftError:
        """To check for project request event and create a message for the root user

        Args:
            project (Project): Project object
            project_event (ProjectEvent): Project event object
            context (AuthedServiceContext): Context of the server

        Returns:
            Union[SyftSuccess, SyftError]: SyftSuccess if message is created else SyftError
        """

        if (
            isinstance(project_event, ProjectRequest)
            and project_event.linked_request.server_uid == context.server.id
        ):
            link = LinkedObject.with_context(project, context=context)
            message = CreateNotification(
                subject=f"A new request has been added to the Project: {project.name}.",
                from_user_verify_key=context.credentials,
                to_user_verify_key=context.server.verify_key,
                linked_obj=link,
            )
            method = context.server.get_service_method(NotificationService.send)
            result = method(context=context, notification=message)
            if isinstance(result, SyftError):
                return result
        return SyftSuccess(message="Successfully Validated Project Request")


TYPE_TO_SERVICE[Project] = ProjectService
SERVICE_TO_TYPES[ProjectService].update({Project})



# stdlib
from enum import Enum


class SyftProtocol(Enum):
    """Enum class to represent the different Syft protocols."""

    HTTP = "http"

    def all(self) -> list:
        return [p.value for p in SyftProtocol]



# stdlib
from concurrent.futures import ThreadPoolExecutor

# third party
from IPython.display import display

# relative
from ..service.dataset.dataset import Dataset
from ..service.metadata.server_metadata import ServerMetadataJSON
from ..service.network.network_service import ServerPeer
from ..service.response import SyftWarning
from ..types.uid import UID
from .client import SyftClient
from .registry import DatasiteRegistry


class SearchResults:
    def __init__(self, results: list[tuple[SyftClient, list[Dataset]]]) -> None:
        self._dataset_client = {}
        self._datasets = []
        for pairs in results:
            client = pairs[0]
            datasets = list(pairs[1])
            for dataset in datasets:
                self._dataset_client[dataset.id] = client
                self._datasets.append(dataset)

    def __getitem__(self, key: int | str | UID) -> Dataset:
        if isinstance(key, int):
            return self._datasets[key]
        else:
            for dataset in self._datasets:
                if isinstance(key, UID):
                    if dataset.id == key:
                        return dataset
                elif isinstance(key, str):
                    if dataset.name == key:
                        return dataset
                    elif str(dataset.id) == key:
                        return dataset
        raise KeyError

    def __repr__(self) -> str:
        return str(self._datasets)

    def _repr_html_(self) -> str:
        return self._datasets._repr_html_()

    def client_for(self, key: Dataset | int | str | UID) -> SyftClient:
        if isinstance(key, Dataset):
            dataset = key
        else:
            dataset = self.__getitem__(key)
        return self._dataset_client[dataset.id]

    def __len__(self) -> int:
        return len(self._datasets)


class Search:
    def __init__(self, datasites: DatasiteRegistry) -> None:
        self.datasites: list[tuple[ServerPeer, ServerMetadataJSON | None]] = (
            datasites.online_datasites
        )

    @staticmethod
    def __search_one_server(
        peer_tuple: tuple[ServerPeer, ServerMetadataJSON], name: str
    ) -> tuple[SyftClient | None, list[Dataset]]:
        try:
            peer, server_metadata = peer_tuple
            client = peer.guest_client
            results = client.api.services.dataset.search(name=name)
            return (client, results)
        except Exception as e:  # noqa
            warning = SyftWarning(
                message=f"Got exception {e} at server {server_metadata.name}"
            )
            display(warning)
            return (None, [])

    def __search(self, name: str) -> list[tuple[SyftClient, list[Dataset]]]:
        with ThreadPoolExecutor(max_workers=20) as executor:
            # results: list[tuple[SyftClient | None, list[Dataset]]] = [
            #     self.__search_one_server(peer_tuple, name) for peer_tuple in self.datasites
            # ]
            results: list[tuple[SyftClient | None, list[Dataset]]] = list(
                executor.map(
                    lambda peer_tuple: self.__search_one_server(peer_tuple, name),
                    self.datasites,
                )
            )
        # filter out SyftError
        filtered = [(client, result) for client, result in results if client and result]

        return filtered

    def search(self, name: str) -> SearchResults:
        """
        Searches for a specific dataset by name.

        Args:
            name (str): The name of the dataset to search for.

        Returns:
            SearchResults: An object containing the search results.
        """
        return SearchResults(self.__search(name))



# stdlib
from collections import defaultdict
import logging
from typing import Any

# third party
from result import Err
from result import Ok
from result import Result

# relative
from ...client.api import ServerIdentity
from ...serde.serializable import serializable
from ...store.document_store import BaseStash
from ...store.document_store import DocumentStore
from ...store.linked_obj import LinkedObject
from ...types.datetime import DateTime
from ...types.syft_object import SyftObject
from ...types.syncable_object import SyncableSyftObject
from ...types.uid import UID
from ...util.telemetry import instrument
from ..action.action_object import ActionObject
from ..action.action_permissions import ActionObjectPermission
from ..action.action_permissions import ActionPermission
from ..action.action_permissions import StoragePermission
from ..api.api import TwinAPIEndpoint
from ..api.api_service import APIService
from ..code.user_code import UserCodeStatusCollection
from ..context import AuthedServiceContext
from ..job.job_stash import Job
from ..response import SyftError
from ..response import SyftSuccess
from ..service import AbstractService
from ..service import TYPE_TO_SERVICE
from ..service import service_method
from ..user.user_roles import ADMIN_ROLE_LEVEL
from .sync_stash import SyncStash
from .sync_state import SyncState

logger = logging.getLogger(__name__)


def get_store(context: AuthedServiceContext, item: SyncableSyftObject) -> Any:
    if isinstance(item, ActionObject):
        service = context.server.get_service("actionservice")  # type: ignore
        return service.store  # type: ignore
    service = context.server.get_service(TYPE_TO_SERVICE[type(item)])  # type: ignore
    return service.stash.partition


@instrument
@serializable(canonical_name="SyncService", version=1)
class SyncService(AbstractService):
    store: DocumentStore
    stash: SyncStash

    def __init__(self, store: DocumentStore):
        self.store = store
        self.stash = SyncStash(store=store)

    def add_actionobject_read_permissions(
        self,
        context: AuthedServiceContext,
        action_object: ActionObject,
        new_permissions: list[ActionObjectPermission],
    ) -> None:
        store_to = context.server.get_service("actionservice").store  # type: ignore
        for permission in new_permissions:
            if permission.permission == ActionPermission.READ:
                store_to.add_permission(permission)

        blob_id = action_object.syft_blob_storage_entry_id
        if blob_id:
            store_to_blob = context.server.get_service(
                "blobstorageservice"
            ).stash.partition  # type: ignore
            for permission in new_permissions:
                if permission.permission == ActionPermission.READ:
                    permission_blob = ActionObjectPermission(
                        uid=blob_id,
                        permission=permission.permission,
                        credentials=permission.credentials,
                    )
                    store_to_blob.add_permission(permission_blob)

    def set_obj_ids(self, context: AuthedServiceContext, x: Any) -> None:
        if hasattr(x, "__dict__") and isinstance(x, SyftObject):
            for val in x.__dict__.values():
                if isinstance(val, list | tuple):
                    for v in val:
                        self.set_obj_ids(context, v)
                elif isinstance(val, dict):
                    for v in val.values():
                        self.set_obj_ids(context, v)
                else:
                    self.set_obj_ids(context, val)
            x.syft_server_location = context.server.id  # type: ignore
            x.syft_client_verify_key = context.credentials
            if hasattr(x, "server_uid"):
                x.server_uid = context.server.id  # type: ignore

    def transform_item(
        self,
        context: AuthedServiceContext,
        item: SyncableSyftObject,
    ) -> SyftObject:
        if isinstance(item, UserCodeStatusCollection):
            identity = ServerIdentity.from_server(context.server)
            res = {}
            for key in item.status_dict.keys():
                # todo, check if they are actually only two servers
                res[identity] = item.status_dict[key]
            item.status_dict = res

        self.set_obj_ids(context, item)
        return item

    def get_stash_for_item(
        self, context: AuthedServiceContext, item: SyftObject
    ) -> BaseStash:
        services = list(context.server.service_path_map.values())  # type: ignore

        all_stashes = {}
        for serv in services:
            if (_stash := getattr(serv, "stash", None)) is not None:
                all_stashes[_stash.object_type] = _stash

        stash = all_stashes.get(type(item), None)
        return stash

    def add_permissions_for_item(
        self,
        context: AuthedServiceContext,
        item: SyftObject,
        new_permissions: list[ActionObjectPermission],
    ) -> None:
        if isinstance(item, ActionObject):
            raise ValueError("ActionObject permissions should be added separately")
        else:
            store = get_store(context, item)  # type: ignore
            for permission in new_permissions:
                if permission.permission == ActionPermission.READ:
                    store.add_permission(permission)

    def add_storage_permissions_for_item(
        self,
        context: AuthedServiceContext,
        item: SyftObject,
        new_permissions: list[StoragePermission],
    ) -> None:
        store = get_store(context, item)
        store.add_storage_permissions(new_permissions)

    def set_object(
        self, context: AuthedServiceContext, item: SyncableSyftObject
    ) -> Result[SyftObject, str]:
        stash = self.get_stash_for_item(context, item)
        creds = context.credentials

        exists = stash.get_by_uid(context.credentials, item.id).ok() is not None

        if isinstance(item, TwinAPIEndpoint):
            # we need the side effect of set function
            # to create an action object
            apiservice: APIService = context.server.get_service("apiservice")  # type: ignore

            res = apiservice.set(context=context, endpoint=item)
            if isinstance(res, SyftError):
                return Err(res.message)
            else:
                return Ok(item)

        if exists:
            res = stash.update(creds, item)
        else:
            # Storage permissions are added separately
            res = stash.set(
                creds,
                item,
                add_storage_permission=False,
            )

        return res

    @service_method(
        path="sync.sync_items",
        name="sync_items",
        roles=ADMIN_ROLE_LEVEL,
    )
    def sync_items(
        self,
        context: AuthedServiceContext,
        items: list[SyncableSyftObject],
        permissions: list[ActionObjectPermission],
        storage_permissions: list[StoragePermission],
        ignored_batches: dict[UID, int],
        unignored_batches: set[UID],
    ) -> SyftSuccess | SyftError:
        permissions_dict = defaultdict(list)
        for permission in permissions:
            permissions_dict[permission.uid].append(permission)

        storage_permissions_dict = defaultdict(list)
        for storage_permission in storage_permissions:
            storage_permissions_dict[storage_permission.uid].append(storage_permission)

        for item in items:
            new_permissions = permissions_dict[item.id.id]
            new_storage_permissions = storage_permissions_dict[item.id.id]
            if isinstance(item, ActionObject):
                self.add_actionobject_read_permissions(context, item, new_permissions)
                self.add_storage_permissions_for_item(
                    context, item, new_storage_permissions
                )
            else:
                item = self.transform_item(context, item)  # type: ignore[unreachable]
                res = self.set_object(context, item)

                if res.is_ok():
                    self.add_permissions_for_item(context, item, new_permissions)
                    self.add_storage_permissions_for_item(
                        context, item, new_storage_permissions
                    )
                else:
                    return SyftError(message=f"Failed to sync {res.err()}")

        res = self.build_current_state(
            context,
            new_items=items,
            new_ignored_batches=ignored_batches,
            new_unignored_batches=unignored_batches,
        )

        if res.is_err():
            return SyftError(message=res.message)
        else:
            new_state = res.ok()
            res = self.stash.set(context.credentials, new_state)
            if res.is_err():
                return SyftError(message=res.message)
            else:
                message = f"Synced {len(items)} items"
                if len(ignored_batches) > 0:
                    message += f", ignored {len(ignored_batches)} batches"
                if len(unignored_batches) > 0:
                    message += f", unignored {len(unignored_batches)} batches"
                return SyftSuccess(message=message)

    @service_method(
        path="sync.get_permissions",
        name="get_permissions",
        roles=ADMIN_ROLE_LEVEL,
    )
    def get_permissions(
        self,
        context: AuthedServiceContext,
        items: list[SyncableSyftObject],
    ) -> tuple[dict[UID, set[str]], dict[UID, set[str]]]:
        permissions = {}
        storage_permissions = {}

        for item in items:
            store = get_store(context, item)
            if store is not None:
                # TODO fix error handling
                uid = item.id.id
                item_permissions = store._get_permissions_for_uid(uid)
                if not item_permissions.is_err():
                    permissions[uid] = item_permissions.ok()

                # TODO fix error handling for storage permissions
                item_storage_permissions = store._get_storage_permissions_for_uid(uid)
                if not item_storage_permissions.is_err():
                    storage_permissions[uid] = item_storage_permissions.ok()
        return permissions, storage_permissions

    def _get_all_items_for_jobs(
        self,
        context: AuthedServiceContext,
    ) -> Result[tuple[list[SyncableSyftObject], dict[UID, str]], str]:
        """
        Returns all Jobs, along with their Logs, ExecutionOutputs and ActionObjects
        """
        items_for_jobs = []
        errors = {}

        job_service = context.server.get_service("jobservice")
        jobs = job_service.get_all(context)
        if isinstance(jobs, SyftError):
            return Err(jobs.message)

        for job in jobs:
            job_items_result = self._get_job_batch(context, job)
            if job_items_result.is_err():
                logger.info(
                    f"Job {job.id} could not be added to SyncState: {job_items_result.err()}"
                )
                errors[job.id] = job_items_result.err()
                continue
            items_for_jobs.extend(job_items_result.ok())

        return Ok((items_for_jobs, errors))

    def _get_job_batch(
        self, context: AuthedServiceContext, job: Job
    ) -> Result[list[SyncableSyftObject], str]:
        job_batch = [job]

        log_service = context.server.get_service("logservice")
        log = log_service.get(context, job.log_id)
        if isinstance(log, SyftError):
            return Err(log.message)
        job_batch.append(log)

        output_service = context.server.get_service("outputservice")
        output = output_service.get_by_job_id(context, job.id)
        if isinstance(output, SyftError):
            return Err(output.message)

        if output is not None:
            job_batch.append(output)
            job_result_ids = set(output.output_id_list)
        else:
            job_result_ids = set()

        if isinstance(job.result, ActionObject):
            job_result_ids.add(job.result.id.id)

        action_service = context.server.get_service("actionservice")
        for result_id in job_result_ids:
            action_object = action_service.get(context, result_id)
            if action_object.is_err():
                return action_object
            job_batch.append(action_object.ok())

        return Ok(job_batch)

    def get_all_syncable_items(
        self, context: AuthedServiceContext
    ) -> Result[tuple[list[SyncableSyftObject], dict[UID, str]], str]:
        all_items: list[SyncableSyftObject] = []

        # NOTE Jobs are handled separately
        services_to_sync = [
            "requestservice",
            "usercodeservice",
            "usercodestatusservice",
            "apiservice",
        ]

        for service_name in services_to_sync:
            service = context.server.get_service(service_name)
            items = service.get_all(context)
            if isinstance(items, SyftError):
                return Err(items.message)
            all_items.extend(items)

        # Gather jobs, logs, outputs and action objects
        items_for_jobs = self._get_all_items_for_jobs(context)
        if items_for_jobs.is_err():
            return items_for_jobs
        items_for_jobs, errors = items_for_jobs.ok()
        all_items.extend(items_for_jobs)

        return Ok((all_items, errors))

    def build_current_state(
        self,
        context: AuthedServiceContext,
        new_items: list[SyncableSyftObject] | None = None,
        new_ignored_batches: dict[UID, int] | None = None,
        new_unignored_batches: set[UID] | None = None,
    ) -> Result[SyncState, str]:
        new_items = new_items if new_items is not None else []
        new_ignored_batches = (
            new_ignored_batches if new_ignored_batches is not None else {}
        )
        unignored_batches: set[UID] = (
            new_unignored_batches if new_unignored_batches is not None else set()
        )
        objects_res = self.get_all_syncable_items(context)
        if objects_res.is_err():
            return objects_res

        objects, errors = objects_res.ok()
        permissions, storage_permissions = self.get_permissions(context, objects)

        previous_state = self.stash.get_latest(context=context)
        if previous_state.is_err():
            return previous_state
        previous_state = previous_state.ok()

        if previous_state is not None:
            previous_state_link = LinkedObject.from_obj(
                obj=previous_state,
                service_type=SyncService,
                server_uid=context.server.id,  # type: ignore
            )
            previous_ignored_batches = previous_state.ignored_batches
        else:
            previous_state_link = None
            previous_ignored_batches = {}

        ignored_batches = {
            **previous_ignored_batches,
            **new_ignored_batches,
        }

        ignored_batches = {
            k: v for k, v in ignored_batches.items() if k not in unignored_batches
        }

        object_sync_dates = (
            previous_state.object_sync_dates.copy() if previous_state else {}
        )
        for obj in new_items:
            object_sync_dates[obj.id.id] = DateTime.now()

        new_state = SyncState(
            server_uid=context.server.id,  # type: ignore
            server_name=context.server.name,  # type: ignore
            server_side_type=context.server.server_side_type,  # type: ignore
            previous_state_link=previous_state_link,
            permissions=permissions,
            storage_permissions=storage_permissions,
            ignored_batches=ignored_batches,
            object_sync_dates=object_sync_dates,
            errors=errors,
        )

        new_state.add_objects(objects, context)

        return Ok(new_state)

    @service_method(
        path="sync._get_state",
        name="_get_state",
        roles=ADMIN_ROLE_LEVEL,
    )
    def _get_state(self, context: AuthedServiceContext) -> SyncState | SyftError:
        res = self.build_current_state(context)
        if res.is_err():
            return SyftError(message=res.value)
        else:
            return res.ok()



# stdlib
from collections.abc import Collection
from collections.abc import Sequence
import logging
from typing import cast

# relative
from ...serde.serializable import serializable
from ...store.document_store import DocumentStore
from ...types.dicttuple import DictTuple
from ...types.uid import UID
from ...util.telemetry import instrument
from ..action.action_permissions import ActionObjectPermission
from ..action.action_permissions import ActionPermission
from ..action.action_service import ActionService
from ..context import AuthedServiceContext
from ..response import SyftError
from ..response import SyftSuccess
from ..service import AbstractService
from ..service import SERVICE_TO_TYPES
from ..service import TYPE_TO_SERVICE
from ..service import service_method
from ..user.user_roles import DATA_OWNER_ROLE_LEVEL
from ..user.user_roles import DATA_SCIENTIST_ROLE_LEVEL
from ..user.user_roles import GUEST_ROLE_LEVEL
from ..warnings import CRUDReminder
from ..warnings import HighSideCRUDWarning
from .dataset import Asset
from .dataset import CreateDataset
from .dataset import Dataset
from .dataset import DatasetPageView
from .dataset import DatasetUpdate
from .dataset_stash import DatasetStash

logger = logging.getLogger(__name__)


def _paginate_collection(
    collection: Collection,
    page_size: int | None = 0,
    page_index: int | None = 0,
) -> slice | None:
    if page_size is None or page_size <= 0:
        return None

    # If chunk size is defined, then split list into evenly sized chunks
    total = len(collection)
    page_index = 0 if page_index is None else page_index

    if page_size > total or page_index >= total // page_size or page_index < 0:
        return None

    start = page_size * page_index
    stop = min(page_size * (page_index + 1), total)
    return slice(start, stop)


def _paginate_dataset_collection(
    datasets: Sequence[Dataset],
    page_size: int | None = 0,
    page_index: int | None = 0,
) -> DictTuple[str, Dataset] | DatasetPageView:
    slice_ = _paginate_collection(datasets, page_size=page_size, page_index=page_index)
    chunk = datasets[slice_] if slice_ is not None else datasets
    results = DictTuple(chunk, lambda dataset: dataset.name)

    return (
        results
        if slice_ is None
        else DatasetPageView(datasets=results, total=len(datasets))
    )


@instrument
@serializable(canonical_name="DatasetService", version=1)
class DatasetService(AbstractService):
    store: DocumentStore
    stash: DatasetStash

    def __init__(self, store: DocumentStore) -> None:
        self.store = store
        self.stash = DatasetStash(store=store)

    @service_method(
        path="dataset.add",
        name="add",
        roles=DATA_OWNER_ROLE_LEVEL,
    )
    def add(
        self, context: AuthedServiceContext, dataset: CreateDataset
    ) -> SyftSuccess | SyftError:
        """Add a Dataset"""
        dataset = dataset.to(Dataset, context=context)
        result = self.stash.set(
            context.credentials,
            dataset,
            add_permissions=[
                ActionObjectPermission(
                    uid=dataset.id, permission=ActionPermission.ALL_READ
                ),
            ],
        )
        if result.is_err():
            return SyftError(message=str(result.err()))
        return SyftSuccess(
            message=f"Dataset uploaded to '{context.server.name}'. "
            f"To see the datasets uploaded by a client on this server, use command `[your_client].datasets`"
        )

    @service_method(
        path="dataset.get_all",
        name="get_all",
        roles=GUEST_ROLE_LEVEL,
        warning=CRUDReminder(),
    )
    def get_all(
        self,
        context: AuthedServiceContext,
        page_size: int | None = 0,
        page_index: int | None = 0,
    ) -> DatasetPageView | DictTuple[str, Dataset] | SyftError:
        """Get a Dataset"""
        result = self.stash.get_all(context.credentials)
        if not result.is_ok():
            return SyftError(message=result.err())

        datasets = result.ok()

        for dataset in datasets:
            if context.server is not None:
                dataset.server_uid = context.server.id
            if dataset.to_be_deleted:
                datasets.remove(dataset)

        return _paginate_dataset_collection(
            datasets=datasets, page_size=page_size, page_index=page_index
        )

    @service_method(path="dataset.search", name="search", roles=GUEST_ROLE_LEVEL)
    def search(
        self,
        context: AuthedServiceContext,
        name: str,
        page_size: int | None = 0,
        page_index: int | None = 0,
    ) -> DatasetPageView | SyftError:
        """Search a Dataset by name"""
        results = self.get_all(context)

        if isinstance(results, SyftError):
            return results

        filtered_results = [
            dataset
            for dataset_name, dataset in results.items()
            if name in dataset_name and not dataset.to_be_deleted
        ]

        return _paginate_dataset_collection(
            filtered_results, page_size=page_size, page_index=page_index
        )

    @service_method(path="dataset.get_by_id", name="get_by_id")
    def get_by_id(self, context: AuthedServiceContext, uid: UID) -> Dataset | SyftError:
        """Get a Dataset"""
        result = self.stash.get_by_uid(context.credentials, uid=uid)
        if result.is_err():
            return SyftError(message=result.err())
        dataset = result.ok()

        if context.server is not None:
            dataset.server_uid = context.server.id
        return dataset

    @service_method(path="dataset.get_by_action_id", name="get_by_action_id")
    def get_by_action_id(
        self, context: AuthedServiceContext, uid: UID
    ) -> list[Dataset] | SyftError:
        """Get Datasets by an Action ID"""
        result = self.stash.search_action_ids(context.credentials, uid=uid)
        if result.is_err():
            return SyftError(message=result.err())
        datasets = result.ok()

        for dataset in datasets:
            if context.server is not None:
                dataset.server_uid = context.server.id

        return datasets

    @service_method(
        path="dataset.get_assets_by_action_id",
        name="get_assets_by_action_id",
        roles=DATA_SCIENTIST_ROLE_LEVEL,
    )
    def get_assets_by_action_id(
        self, context: AuthedServiceContext, uid: UID
    ) -> list[Asset] | SyftError:
        """Get Assets by an Action ID"""
        datasets = self.get_by_action_id(context=context, uid=uid)
        if isinstance(datasets, SyftError):
            return datasets
        for dataset in datasets:
            if dataset.to_be_deleted:
                datasets.remove(dataset)
        return [
            asset
            for dataset in datasets
            for asset in dataset.asset_list
            if asset.action_id == uid
        ]

    @service_method(
        path="dataset.delete",
        name="delete",
        roles=DATA_OWNER_ROLE_LEVEL,
        warning=HighSideCRUDWarning(confirmation=True),
    )
    def delete(
        self, context: AuthedServiceContext, uid: UID, delete_assets: bool = True
    ) -> SyftSuccess | SyftError:
        """
        Soft delete: keep the dataset object, only remove the blob store entries
        After soft deleting a dataset, the user will not be able to
        see it using the `datasets.get_all` endpoint.
        Delete unique `dataset.name` key and leave UID, just rename it in case the
        user upload a new dataset with the same name.
        """
        # check if the dataset exists
        dataset = self.get_by_id(context=context, uid=uid)
        if isinstance(dataset, SyftError):
            return dataset

        return_msg = []
        if delete_assets:
            # delete the dataset's assets
            for asset in dataset.asset_list:
                msg = (
                    f"ActionObject {asset.action_id} "
                    f"linked with Assset {asset.id} "
                    f"in Dataset {uid}"
                )

                action_service = cast(
                    ActionService, context.server.get_service(ActionService)
                )
                del_res: SyftSuccess | SyftError = action_service.delete(
                    context=context, uid=asset.action_id, soft_delete=True
                )

                if isinstance(del_res, SyftError):
                    del_msg = f"Failed to delete {msg}: {del_res.message}"
                    logger.error(del_msg)
                    return del_res

                logger.info(f"Successfully deleted {msg}: {del_res.message}")

                return_msg.append(f"Asset with id '{asset.id}' successfully deleted.")

        # soft delete the dataset object from the store
        dataset_update = DatasetUpdate(
            id=uid, name=f"_deleted_{dataset.name}_{uid}", to_be_deleted=True
        )
        result = self.stash.update(context.credentials, dataset_update)
        if result.is_err():
            return SyftError(message=result.err())
        return_msg.append(f"Dataset with id '{uid}' successfully deleted.")
        return SyftSuccess(message="\n".join(return_msg))


TYPE_TO_SERVICE[Dataset] = DatasetService
SERVICE_TO_TYPES[DatasetService].update({Dataset})



# stdlib
from collections.abc import Callable

# third party
import requests

# relative
from ...serde.serializable import serializable
from ...store.document_store import DocumentStore
from ...util.util import str_to_bool
from ..context import AuthedServiceContext
from ..response import SyftError
from ..response import SyftSuccess
from ..service import AbstractService
from ..service import service_method
from ..user.user_roles import GUEST_ROLE_LEVEL
from .attestation_constants import ATTESTATION_SERVICE_URL
from .attestation_constants import ATTEST_CPU_ENDPOINT
from .attestation_constants import ATTEST_GPU_ENDPOINT


@serializable(canonical_name="AttestationService", version=1)
class AttestationService(AbstractService):
    """This service is responsible for getting all sorts of attestations for any client."""

    def __init__(self, store: DocumentStore) -> None:
        self.store = store

    def perform_request(
        self, method: Callable, endpoint: str, raw: bool = False
    ) -> SyftSuccess | SyftError | str:
        try:
            response = method(f"{ATTESTATION_SERVICE_URL}{endpoint}")
            response.raise_for_status()
            message = response.json().get("result")
            raw_token = response.json().get("token")
            if raw:
                return raw_token
            elif str_to_bool(message):
                return SyftSuccess(message=message)
            else:
                return SyftError(message=message)
        except requests.HTTPError:
            return SyftError(message=f"{response.json()['detail']}")
        except requests.RequestException as e:
            return SyftError(message=f"Failed to perform request. {e}")

    @service_method(
        path="attestation.get_cpu_attestation",
        name="get_cpu_attestation",
        roles=GUEST_ROLE_LEVEL,
    )
    def get_cpu_attestation(
        self, context: AuthedServiceContext, raw_token: bool = False
    ) -> str | SyftError | SyftSuccess:
        return self.perform_request(requests.get, ATTEST_CPU_ENDPOINT, raw_token)

    @service_method(
        path="attestation.get_gpu_attestation",
        name="get_gpu_attestation",
        roles=GUEST_ROLE_LEVEL,
    )
    def get_gpu_attestation(
        self, context: AuthedServiceContext, raw_token: bool = False
    ) -> str | SyftError | SyftSuccess:
        return self.perform_request(requests.get, ATTEST_GPU_ENDPOINT, raw_token)



# future
from __future__ import annotations

# stdlib
from collections import defaultdict
from copy import deepcopy
import logging
from pathlib import Path
import sqlite3
import tempfile
from typing import Any

# third party
from pydantic import Field
from pydantic import field_validator
from result import Err
from result import Ok
from result import Result
from typing_extensions import Self

# relative
from ..serde.deserialize import _deserialize
from ..serde.serializable import serializable
from ..serde.serialize import _serialize
from ..types.uid import UID
from ..util.util import thread_ident
from .document_store import DocumentStore
from .document_store import PartitionSettings
from .document_store import StoreClientConfig
from .document_store import StoreConfig
from .kv_document_store import KeyValueBackingStore
from .kv_document_store import KeyValueStorePartition
from .locks import LockingConfig
from .locks import NoLockingConfig
from .locks import SyftLock

logger = logging.getLogger(__name__)

# here we can create a single connection per cache_key
# since pytest is concurrent processes, we need to isolate each connection
# by its filename and optionally the thread that its running in
# we keep track of each SQLiteBackingStore init in REF_COUNTS
# when it hits 0 we can close the connection and release the file descriptor
SQLITE_CONNECTION_POOL_DB: dict[str, sqlite3.Connection] = {}
SQLITE_CONNECTION_POOL_CUR: dict[str, sqlite3.Cursor] = {}
REF_COUNTS: dict[str, int] = defaultdict(int)


def cache_key(db_name: str) -> str:
    return f"{db_name}_{thread_ident()}"


def _repr_debug_(value: Any) -> str:
    if hasattr(value, "_repr_debug_"):
        return str(value._repr_debug_())
    return repr(value)


def raise_exception(table_name: str, e: Exception) -> None:
    if "disk I/O error" in str(e):
        message = f"Error usually related to concurrent writes. {str(e)}"
        raise Exception(message)

    if "Cannot operate on a closed database" in str(e):
        message = (
            "Error usually related to calling self.db.close()"
            + f"before last SQLiteBackingStore.__del__ gets called. {str(e)}"
        )
        raise Exception(message)

    # if its something else other than "table already exists" raise original e
    if f"table {table_name} already exists" not in str(e):
        raise e


@serializable(
    attrs=["index_name", "settings", "store_config"],
    canonical_name="SQLiteBackingStore",
    version=1,
)
class SQLiteBackingStore(KeyValueBackingStore):
    """Core Store logic for the SQLite stores.

    Parameters:
        `index_name`: str
            Index name
        `settings`: PartitionSettings
            Syft specific settings
        `store_config`: SQLiteStoreConfig
            Connection Configuration
        `ddtype`: Type
            Class used as fallback on `get` errors
    """

    def __init__(
        self,
        index_name: str,
        settings: PartitionSettings,
        store_config: StoreConfig,
        ddtype: type | None = None,
    ) -> None:
        self.index_name = index_name
        self.settings = settings
        self.store_config = store_config
        self._ddtype = ddtype
        if self.store_config.client_config:
            self.file_path = self.store_config.client_config.file_path
        if store_config.client_config:
            self.db_filename = store_config.client_config.filename

        self.lock = SyftLock(NoLockingConfig())
        self.create_table()
        REF_COUNTS[cache_key(self.db_filename)] += 1

    @property
    def table_name(self) -> str:
        return f"{self.settings.name}_{self.index_name}"

    def _connect(self) -> None:
        # SQLite is not thread safe by default so we ensure that each connection
        # comes from a different thread. In cases of Uvicorn and other AWSGI servers
        # there will be many threads handling incoming requests so we need to ensure
        # that different connections are used in each thread. By using a dict for the
        # _db and _cur we can ensure they are never shared

        path = Path(self.file_path)
        if not path.exists():
            path.parent.mkdir(parents=True, exist_ok=True)

        if self.store_config.client_config:
            connection = sqlite3.connect(
                self.file_path,
                timeout=self.store_config.client_config.timeout,
                check_same_thread=False,  # do we need this if we use the lock?
                # check_same_thread=self.store_config.client_config.check_same_thread,
            )
            # Set journal mode to WAL.
            connection.execute("PRAGMA journal_mode = WAL")
            connection.execute("PRAGMA busy_timeout = 5000")
            connection.execute("PRAGMA temp_store = 2")
            connection.execute("PRAGMA synchronous = 1")
            SQLITE_CONNECTION_POOL_DB[cache_key(self.db_filename)] = connection

    def create_table(self) -> None:
        try:
            with self.lock:
                self.cur.execute(
                    f"create table {self.table_name} (uid VARCHAR(32) NOT NULL PRIMARY KEY, "  # nosec
                    + "repr TEXT NOT NULL, value BLOB NOT NULL, "  # nosec
                    + "sqltime TIMESTAMP DEFAULT CURRENT_TIMESTAMP NOT NULL)"  # nosec
                )
                self.db.commit()
        except Exception as e:
            raise_exception(self.table_name, e)

    @property
    def db(self) -> sqlite3.Connection:
        if cache_key(self.db_filename) not in SQLITE_CONNECTION_POOL_DB:
            self._connect()
        return SQLITE_CONNECTION_POOL_DB[cache_key(self.db_filename)]

    @property
    def cur(self) -> sqlite3.Cursor:
        if cache_key(self.db_filename) not in SQLITE_CONNECTION_POOL_CUR:
            SQLITE_CONNECTION_POOL_CUR[cache_key(self.db_filename)] = self.db.cursor()

        return SQLITE_CONNECTION_POOL_CUR[cache_key(self.db_filename)]

    def _close(self) -> None:
        self._commit()
        REF_COUNTS[cache_key(self.db_filename)] -= 1
        if REF_COUNTS[cache_key(self.db_filename)] <= 0:
            # once you close it seems like other object references can't re-use the
            # same connection

            self.db.close()
            db_key = cache_key(self.db_filename)
            if db_key in SQLITE_CONNECTION_POOL_CUR:
                # NOTE if we don't remove the cursor, the cursor cache_key can clash with a future thread id
                del SQLITE_CONNECTION_POOL_CUR[db_key]
            del SQLITE_CONNECTION_POOL_DB[cache_key(self.db_filename)]
        else:
            # don't close yet because another SQLiteBackingStore is probably still open
            pass

    def _commit(self) -> None:
        self.db.commit()

    def _execute(
        self, sql: str, *args: list[Any] | None
    ) -> Result[Ok[sqlite3.Cursor], Err[str]]:
        with self.lock:
            cursor: sqlite3.Cursor | None = None
            # err = None
            try:
                cursor = self.cur.execute(sql, *args)
            except Exception as e:
                raise_exception(self.table_name, e)

            # TODO: Which exception is safe to rollback on?
            # we should map out some more clear exceptions that can be returned
            # rather than halting the program like disk I/O error etc
            # self.db.rollback()  # Roll back all changes if an exception occurs.
            # err = Err(str(e))
            self.db.commit()  # Commit if everything went ok

            # if err is not None:
            #     return err

            return Ok(cursor)

    def _set(self, key: UID, value: Any) -> None:
        if self._exists(key):
            self._update(key, value)
        else:
            insert_sql = (
                f"insert into {self.table_name} (uid, repr, value) VALUES (?, ?, ?)"  # nosec
            )
            data = _serialize(value, to_bytes=True)
            res = self._execute(insert_sql, [str(key), _repr_debug_(value), data])
            if res.is_err():
                raise ValueError(res.err())

    def _update(self, key: UID, value: Any) -> None:
        insert_sql = (
            f"update {self.table_name} set uid = ?, repr = ?, value = ? where uid = ?"  # nosec
        )
        data = _serialize(value, to_bytes=True)
        res = self._execute(insert_sql, [str(key), _repr_debug_(value), data, str(key)])
        if res.is_err():
            raise ValueError(res.err())

    def _get(self, key: UID) -> Any:
        select_sql = f"select * from {self.table_name} where uid = ? order by sqltime"  # nosec
        res = self._execute(select_sql, [str(key)])
        if res.is_err():
            raise KeyError(f"Query {select_sql} failed")
        cursor = res.ok()

        row = cursor.fetchone()
        if row is None or len(row) == 0:
            raise KeyError(f"{key} not in {type(self)}")
        data = row[2]
        return _deserialize(data, from_bytes=True)

    def _exists(self, key: UID) -> bool:
        select_sql = f"select uid from {self.table_name} where uid = ?"  # nosec

        res = self._execute(select_sql, [str(key)])
        if res.is_err():
            return False
        cursor = res.ok()

        row = cursor.fetchone()
        if row is None:
            return False

        return bool(row)

    def _get_all(self) -> Any:
        select_sql = f"select * from {self.table_name} order by sqltime"  # nosec
        keys = []
        data = []

        res = self._execute(select_sql)
        if res.is_err():
            return {}
        cursor = res.ok()

        rows = cursor.fetchall()
        if rows is None:
            return {}

        for row in rows:
            keys.append(UID(row[0]))
            data.append(_deserialize(row[2], from_bytes=True))
        return dict(zip(keys, data))

    def _get_all_keys(self) -> Any:
        select_sql = f"select uid from {self.table_name} order by sqltime"  # nosec

        res = self._execute(select_sql)
        if res.is_err():
            return []
        cursor = res.ok()

        rows = cursor.fetchall()
        if rows is None:
            return []

        keys = [UID(row[0]) for row in rows]
        return keys

    def _delete(self, key: UID) -> None:
        select_sql = f"delete from {self.table_name} where uid = ?"  # nosec
        res = self._execute(select_sql, [str(key)])
        if res.is_err():
            raise ValueError(res.err())

    def _delete_all(self) -> None:
        select_sql = f"delete from {self.table_name}"  # nosec
        res = self._execute(select_sql)
        if res.is_err():
            raise ValueError(res.err())

    def _len(self) -> int:
        select_sql = f"select count(uid) from {self.table_name}"  # nosec
        res = self._execute(select_sql)
        if res.is_err():
            raise ValueError(res.err())
        cursor = res.ok()
        cnt = cursor.fetchone()[0]
        return cnt

    def __setitem__(self, key: Any, value: Any) -> None:
        self._set(key, value)

    def __getitem__(self, key: Any) -> Self:
        try:
            return self._get(key)
        except KeyError as e:
            if self._ddtype is not None:
                return self._ddtype()
            raise e

    def __repr__(self) -> str:
        return repr(self._get_all())

    def __len__(self) -> int:
        return self._len()

    def __delitem__(self, key: str) -> None:
        self._delete(key)

    def clear(self) -> None:
        self._delete_all()

    def copy(self) -> Self:
        return deepcopy(self)

    def keys(self) -> Any:
        return self._get_all_keys()

    def values(self) -> Any:
        return self._get_all().values()

    def items(self) -> Any:
        return self._get_all().items()

    def pop(self, key: Any) -> Self:
        value = self._get(key)
        self._delete(key)
        return value

    def __contains__(self, key: Any) -> bool:
        return self._exists(key)

    def __iter__(self) -> Any:
        return iter(self.keys())

    def __del__(self) -> None:
        try:
            self._close()
        except Exception as e:
            logger.error("Could not close connection", exc_info=e)


@serializable(canonical_name="SQLiteStorePartition", version=1)
class SQLiteStorePartition(KeyValueStorePartition):
    """SQLite StorePartition

    Parameters:
        `settings`: PartitionSettings
            PySyft specific settings, used for indexing and partitioning
        `store_config`: SQLiteStoreConfig
            SQLite specific configuration
    """

    def close(self) -> None:
        self.lock.acquire()
        try:
            # I think we don't want these now, because of the REF_COUNT?
            # self.data._close()
            # self.unique_keys._close()
            # self.searchable_keys._close()
            pass
        except BaseException:
            pass
        self.lock.release()

    def commit(self) -> None:
        self.lock.acquire()
        try:
            self.data._commit()
            self.unique_keys._commit()
            self.searchable_keys._commit()
        except BaseException:
            pass
        self.lock.release()


# the base document store is already a dict but we can change it later
@serializable(canonical_name="SQLiteDocumentStore", version=1)
class SQLiteDocumentStore(DocumentStore):
    """SQLite Document Store

    Parameters:
        `store_config`: StoreConfig
            SQLite specific configuration, including connection details and client class type.
    """

    partition_type = SQLiteStorePartition


@serializable(canonical_name="SQLiteStoreClientConfig", version=1)
class SQLiteStoreClientConfig(StoreClientConfig):
    """SQLite connection config

    Parameters:
        `filename` : str
            Database name
        `path` : Path or str
            Database folder
        `check_same_thread`: bool
            If True (default), ProgrammingError will be raised if the database connection is used
            by a thread other than the one that created it. If False, the connection may be accessed
            in multiple threads; write operations may need to be serialized by the user to avoid
            data corruption.
        `timeout`: int
            How many seconds the connection should wait before raising an exception, if the database
            is locked by another connection. If another connection opens a transaction to modify the
            database, it will be locked until that transaction is committed. Default five seconds.
    """

    filename: str = "syftdb.sqlite"
    path: str | Path = Field(default_factory=tempfile.gettempdir)
    check_same_thread: bool = True
    timeout: int = 5

    # We need this in addition to Field(default_factory=...)
    # so users can still do SQLiteStoreClientConfig(path=None)
    @field_validator("path", mode="before")
    @classmethod
    def __default_path(cls, path: str | Path | None) -> str | Path:
        if path is None:
            return tempfile.gettempdir()
        return path

    @property
    def file_path(self) -> Path | None:
        return Path(self.path) / self.filename


@serializable()
class SQLiteStoreConfig(StoreConfig):
    __canonical_name__ = "SQLiteStoreConfig"
    """SQLite Store config, used by SQLiteStorePartition

    Parameters:
        `client_config`: SQLiteStoreClientConfig
            SQLite connection configuration
        `store_type`: DocumentStore
            Class interacting with QueueStash. Default: SQLiteDocumentStore
        `backing_store`: KeyValueBackingStore
            The Store core logic. Default: SQLiteBackingStore
        locking_config: LockingConfig
            The config used for store locking. Available options:
                * NoLockingConfig: no locking, ideal for single-thread stores.
                * ThreadingLockingConfig: threading-based locking, ideal for same-process in-memory stores.
            Defaults to NoLockingConfig.
    """

    client_config: SQLiteStoreClientConfig
    store_type: type[DocumentStore] = SQLiteDocumentStore
    backing_store: type[KeyValueBackingStore] = SQLiteBackingStore
    locking_config: LockingConfig = Field(default_factory=NoLockingConfig)



# stdlib
from enum import Enum
from enum import auto

# relative
from ...serde.serializable import serializable


@serializable(canonical_name="NOTIFIERS", version=1)
class NOTIFIERS(Enum):
    EMAIL = auto()
    SMS = auto()
    SLACK = auto()
    APP = auto()



# stdlib
from enum import Enum
from typing import Any
from typing import cast

# third party
import docker
from docker.models.containers import Container

# relative
from ...client.api import APIRegistry
from ...serde.serializable import serializable
from ...store.linked_obj import LinkedObject
from ...types.base import SyftBaseModel
from ...types.datetime import DateTime
from ...types.syft_object import SYFT_OBJECT_VERSION_1
from ...types.syft_object import SyftObject
from ...types.syft_object import short_uid
from ...types.uid import UID
from ...util import options
from ...util.colors import SURFACE
from ...util.notebook_ui.styles import FONT_CSS
from ...util.notebook_ui.styles import ITABLES_CSS
from ..response import SyftError
from .worker_image import SyftWorkerImage


@serializable(canonical_name="WorkerStatus", version=1)
class WorkerStatus(Enum):
    PENDING = "Pending"
    RUNNING = "Running"
    STOPPED = "Stopped"
    RESTARTED = "Restarted"


@serializable(canonical_name="ConsumerState", version=1)
class ConsumerState(Enum):
    IDLE = "Idle"
    CONSUMING = "Consuming"
    DETACHED = "Detached"


@serializable(canonical_name="WorkerHealth", version=1)
class WorkerHealth(Enum):
    HEALTHY = "✅"
    UNHEALTHY = "❌"


@serializable()
class SyftWorker(SyftObject):
    __canonical_name__ = "SyftWorker"
    __version__ = SYFT_OBJECT_VERSION_1

    __attr_unique__ = ["name"]
    __attr_searchable__ = ["name", "container_id", "to_be_deleted"]
    __repr_attrs__ = [
        "name",
        "container_id",
        "image",
        "status",
        "healthcheck",
        "worker_pool_name",
        "created_at",
    ]

    id: UID
    name: str
    container_id: str | None = None
    created_at: DateTime = DateTime.now()
    healthcheck: WorkerHealth | None = None
    status: WorkerStatus
    image: SyftWorkerImage | None = None
    worker_pool_name: str
    consumer_state: ConsumerState = ConsumerState.DETACHED
    job_id: UID | None = None
    to_be_deleted: bool = False

    @property
    def logs(self) -> str | SyftError:
        api = APIRegistry.api_for(
            server_uid=self.syft_server_location,
            user_verify_key=self.syft_client_verify_key,
        )
        if api is None:
            return SyftError(message=f"You must login to {self.server_uid}")
        return api.services.worker.logs(uid=self.id)

    def get_job_repr(self) -> str:
        if self.job_id is not None:
            api = APIRegistry.api_for(
                server_uid=self.syft_server_location,
                user_verify_key=self.syft_client_verify_key,
            )
            if api is None:
                return SyftError(message=f"You must login to {self.server_uid}")
            job = api.services.job.get(self.job_id)
            if job.action.user_code_id is not None:
                func_name = api.services.code.get_by_id(
                    job.action.user_code_id
                ).service_func_name
                return f"{func_name} ({short_uid(self.job_id)})"
            else:
                return f"action ({short_uid(self.job_id)})"
        else:
            return ""

    def refresh_status(self) -> SyftError | None:
        api = APIRegistry.api_for(
            server_uid=self.syft_server_location,
            user_verify_key=self.syft_client_verify_key,
        )
        if api is None:
            return SyftError(message=f"You must login to {self.server_uid}")

        res = api.services.worker.status(uid=self.id)
        if isinstance(res, SyftError):
            return res

        self.status, self.healthcheck = res
        return None

    def _coll_repr_(self) -> dict[str, Any]:
        self.refresh_status()

        if self.image and self.image.image_identifier:
            image_name_with_tag = self.image.image_identifier.full_name_with_tag
        else:
            image_name_with_tag = "In Memory Worker"

        healthcheck = self.healthcheck.value if self.healthcheck is not None else ""

        return {
            "Name": self.name,
            "Image": image_name_with_tag,
            "Healthcheck (health / unhealthy)": f"{healthcheck}",
            "Status": f"{self.status.value}",
            "Job": self.get_job_repr(),
            "Created at": str(self.created_at),
            "Container id": self.container_id,
            "Consumer state": str(self.consumer_state.value.lower()),
        }


@serializable()
class WorkerPool(SyftObject):
    __canonical_name__ = "WorkerPool"
    __version__ = SYFT_OBJECT_VERSION_1

    __attr_unique__ = ["name"]
    __attr_searchable__ = ["name", "image_id"]
    __repr_attrs__ = [
        "name",
        "image",
        "max_count",
        "workers",
        "created_at",
    ]
    __table_sort_attr__ = "Created at"

    name: str
    image_id: UID | None = None
    max_count: int
    worker_list: list[LinkedObject]
    created_at: DateTime = DateTime.now()

    @property
    def image(self) -> SyftWorkerImage | SyftError | None:
        """
        Get the pool's image using the worker_image service API. This way we
        get the latest state of the image from the SyftWorkerImageStash
        """
        api = APIRegistry.api_for(
            server_uid=self.syft_server_location,
            user_verify_key=self.syft_client_verify_key,
        )
        if api is not None and api.services is not None:
            return api.services.worker_image.get_by_uid(uid=self.image_id)
        else:
            return None

    @property
    def running_workers(self) -> list[SyftWorker] | SyftError:
        """Query the running workers using an API call to the server"""
        _running_workers = [
            worker for worker in self.workers if worker.status == WorkerStatus.RUNNING
        ]

        return _running_workers

    @property
    def healthy_workers(self) -> list[SyftWorker] | SyftError:
        """
        Query the healthy workers using an API call to the server
        """
        _healthy_workers = [
            worker
            for worker in self.workers
            if worker.healthcheck == WorkerHealth.HEALTHY
        ]

        return _healthy_workers

    def _coll_repr_(self) -> dict[str, Any]:
        if self.image and self.image.image_identifier:
            image_name_with_tag = self.image.image_identifier.full_name_with_tag
        else:
            image_name_with_tag = "In Memory Worker"
        return {
            "Pool Name": self.name,
            "Workers": len(self.workers),
            "Healthy (healthy / all)": f"{len(self.healthy_workers)} / {self.max_count}",
            "Running (running / all)": f"{len(self.running_workers)} / {self.max_count}",
            "Image": image_name_with_tag,
            "Created at": str(self.created_at),
        }

    def _repr_html_(self) -> Any:
        return f"""
            <style>
            {FONT_CSS}
            .syft-dataset {{color: {SURFACE[options.color_theme]};}}
            .syft-dataset h3,
            .syft-dataset p
              {{font-family: 'Open Sans';}}
              {ITABLES_CSS}
            </style>
            <div class='syft-dataset'>
            <h3>{self.name}</h3>
            <p class='paragraph-sm'>
                <strong><span class='pr-8'>Created on: </span></strong>
                {self.created_at}
            </p>
            <p class='paragraph-sm'>
                <strong><span class='pr-8'>Healthy Workers:</span></strong>
                {len(self.healthy_workers)} / {self.max_count}
            </p>
            <p class='paragraph-sm'>
                <strong><span class='pr-8'>Running Workers:</span></strong>
                {len(self.running_workers)} / {self.max_count}
            </p>
            {self.workers._repr_html_()}
            """

    @property
    def workers(self) -> list[SyftWorker]:
        resolved_workers = []
        for worker in self.worker_list:
            resolved_worker = worker.resolve
            if isinstance(resolved_worker, SyftError) or resolved_worker is None:
                continue
            resolved_worker.refresh_status()
            resolved_workers.append(resolved_worker)
        return resolved_workers


@serializable(canonical_name="WorkerOrchestrationType", version=1)
class WorkerOrchestrationType(Enum):
    DOCKER = "docker"
    KUBERNETES = "k8s"
    PYTHON = "python"


@serializable(canonical_name="ContainerSpawnStatus", version=1)
class ContainerSpawnStatus(SyftBaseModel):
    __repr_attrs__ = ["worker_name", "worker", "error"]

    worker_name: str
    worker: SyftWorker | None = None
    error: str | None = None


def _get_worker_container(
    client: docker.DockerClient,
    worker: SyftWorker,
) -> Container | SyftError:
    try:
        return cast(Container, client.containers.get(worker.container_id))
    except docker.errors.NotFound as e:
        return SyftError(message=f"Worker {worker.id} container not found. Error {e}")
    except docker.errors.APIError as e:
        return SyftError(
            message=f"Unable to access worker {worker.id} container. "
            + f"Container server error {e}"
        )


_CONTAINER_STATUS_TO_WORKER_STATUS: dict[str, WorkerStatus] = dict(
    [
        ("running", WorkerStatus.RUNNING),
        *(
            (status, WorkerStatus.STOPPED)
            for status in ["paused", "removing", "exited", "dead"]
        ),
        ("restarting", WorkerStatus.RESTARTED),
        ("created", WorkerStatus.PENDING),
    ]
)


def _get_worker_container_status(
    client: docker.DockerClient,
    worker: SyftWorker,
    container: Container | None = None,
) -> Container | SyftError:
    if container is None:
        container = _get_worker_container(client, worker)

    if isinstance(container, SyftError):
        return container

    container_status = container.status

    return _CONTAINER_STATUS_TO_WORKER_STATUS.get(
        container_status,
        SyftError(message=f"Unknown container status: {container_status}"),
    )



# stdlib
from collections.abc import Callable
from enum import Enum
from enum import auto
from functools import partial
import html
import secrets
from typing import Any
from uuid import uuid4

# third party
from IPython import display
import ipywidgets as widgets
from ipywidgets import Button
from ipywidgets import Checkbox
from ipywidgets import HBox
from ipywidgets import HTML
from ipywidgets import Layout
from ipywidgets import VBox

# relative
from ...client.sync_decision import SyncDirection
from ...types.uid import UID
from ...util.notebook_ui.components.sync import Alert
from ...util.notebook_ui.components.sync import CopyIDButton
from ...util.notebook_ui.components.sync import MainDescription
from ...util.notebook_ui.components.sync import SyncWidgetHeader
from ...util.notebook_ui.components.sync import TypeLabel
from ...util.notebook_ui.components.tabulator_template import build_tabulator_table
from ...util.notebook_ui.components.tabulator_template import highlight_single_row
from ...util.notebook_ui.components.tabulator_template import update_table_cell
from ...util.notebook_ui.styles import CSS_CODE
from ..action.action_object import ActionObject
from ..api.api import TwinAPIEndpoint
from ..log.log import SyftLog
from ..response import SyftError
from ..response import SyftSuccess
from .diff_state import ObjectDiff
from .diff_state import ObjectDiffBatch

# Standard div Jupyter Lab uses for notebook outputs
# This is needed to use alert styles from SyftSuccess and SyftError
NOTEBOOK_OUTPUT_DIV = """
<div class="lm-Widget
            jp-RenderedHTMLCommon
            jp-RenderedHTML
            jp-mod-trusted
            jp-OutputArea-output"
     data-mime-type="text/html">
    {content}<br>
</div>
"""


class DiffStatus(Enum):
    NEW = auto()
    SAME = auto()
    MODIFIED = auto()
    DELETED = auto()


background_colors = {
    DiffStatus.NEW: "#D5F1D5;",
    DiffStatus.SAME: "transparent",
    DiffStatus.MODIFIED: "#FEE9CD;",
    DiffStatus.DELETED: "#ffdddd;",
}


colors = {
    DiffStatus.NEW: "#256B24;",
    DiffStatus.SAME: "#353243",
    DiffStatus.MODIFIED: "#B8520A;",
    DiffStatus.DELETED: "#353243",
}


def create_diff_html(
    title: str,
    properties: dict[str, str],
    statuses: dict[str, DiffStatus],
) -> str:
    html_str = f"<div style='width: 100%;'>{title}<br>"
    html_str += "<div style='font-family: monospace; border-left: 1px solid #B4B0BF; padding-left: 10px;'>"

    for attr, val in properties.items():
        status = statuses[attr]
        val = val if val is not None else ""
        style = f"background-color: {background_colors[status]}; color: {colors[status]}; display: block; white-space: pre-wrap; margin-bottom: 5px;"  # noqa: E501
        content = html.escape(f"{attr}: {val}")
        html_str += f"<div style='{style}'>{content}</div>"

    html_str += "</div></div>"

    return html_str


# TODO move CSS/HTML/JS outside function


class MainObjectDiffWidget:
    def __init__(
        self,
        diff: ObjectDiff,
        direction: SyncDirection,
        with_box: bool = True,
        show_share_warning: bool = False,
        build_state: bool = True,
    ):
        build_state = build_state

        if build_state:
            self.low_properties = diff.repr_attr_dict("low")
            self.high_properties = diff.repr_attr_dict("high")
            self.statuses = diff.repr_attr_diffstatus_dict()
        else:
            self.low_properties = {}
            self.high_properties = {}
            self.statuses = {}

        self.direction = direction
        self.diff: ObjectDiff = diff
        self.with_box = with_box
        self.show_share_warning = show_share_warning
        self.sync = True
        self.is_main_widget: bool = True

        self.widget = self.build()

    def set_share_private_data(self) -> None:
        # No-op for main widget
        pass

    @property
    def mockify(self) -> bool:
        return not self.share_private_data

    @property
    def share_private_data(self) -> bool:
        # there are TwinAPIEndpoint.__private_sync_attr_mocks__
        return not isinstance(self.diff.non_empty_object, TwinAPIEndpoint)

    @property
    def warning_html(self) -> str:
        if isinstance(self.diff.non_empty_object, TwinAPIEndpoint):
            message = "Only the public function of a TwinAPI will be synced to the low-side server."
            return Alert(message=message).to_html()
        elif self.show_share_warning:
            message = (
                "By default only the object wrapper will be synced. "
                "If you would like to sync the real data please "
                'activate the "Sync Real Data" button above.'
            )
            return Alert(message=message).to_html()
        else:
            return ""

    def build(self) -> widgets.HBox:
        all_keys = list(self.low_properties.keys()) + list(self.high_properties.keys())
        low_properties = {}
        high_properties = {}
        for k in all_keys:
            low_properties[k] = self.low_properties.get(k, None)
            high_properties[k] = self.high_properties.get(k, None)

        if self.direction == SyncDirection.LOW_TO_HIGH:
            from_properties = low_properties
            to_properties = high_properties
            source_side = "Low side"
            target_side = "High side"
        else:
            from_properties = high_properties
            to_properties = low_properties
            source_side = "High side"
            target_side = "Low side"

        html_from = create_diff_html(
            f"From <i>{source_side}</i> (new values)", from_properties, self.statuses
        )
        html_to = create_diff_html(
            f"To <i>{target_side}</i> (old values)", to_properties, self.statuses
        )

        widget_from = widgets.HTML(
            value=html_from, layout=widgets.Layout(width="50%", overflow="auto")
        )
        widget_to = widgets.HTML(
            value=html_to, layout=widgets.Layout(width="50%", overflow="auto")
        )
        css_accordion = """
            <style>
            .diff-container {
                border: 0.5px solid #B4B0BF;
            }
            </style>
        """

        result = widgets.HBox([HTML(css_accordion), widget_from, widget_to])

        warning = self.warning_html
        if warning:
            result = VBox([widgets.HTML(warning), result])

        if self.with_box:
            result._dom_classes = result._dom_classes + ("diff-container",)

        return result


class CollapsableObjectDiffWidget:
    def __init__(
        self,
        diff: ObjectDiff,
        direction: SyncDirection,
        build_state: bool = True,
    ):
        self.direction = direction
        self.build_state = build_state
        self.share_private_data = False
        self.diff: ObjectDiff = diff
        self.sync: bool = False
        self.is_main_widget: bool = False
        self.has_private_data = isinstance(
            self.diff.non_empty_object, SyftLog | ActionObject | TwinAPIEndpoint
        )
        self.widget = self.build()
        self.set_and_disable_sync()

    @property
    def mockify(self) -> bool:
        if self.has_private_data and not self.share_private_data:
            return True
        else:
            return False

    @property
    def warning_html(self) -> str:
        if self.show_share_button:
            message = (
                "By default only the object wrapper will be synced. "
                "If you would like to sync the real log data please "
                "activate the “Real Data” button above."
            )
            return Alert(message=message).to_html()
        return ""

    @property
    def show_share_button(self) -> bool:
        return self.has_private_data

    @property
    def title(self) -> str:
        object = self.diff.non_empty_object
        if object is None:
            return "n/a"
        type_html = TypeLabel(object=object).to_html()
        description_html = MainDescription(object=object).to_html()
        copy_id_button = CopyIDButton(copy_text=str(object.id.id), max_width=60)

        second_line_html = f"""
            <div class="widget-header2">
            <div class="widget-header2-2">
            {type_html} {description_html}
            </div>
            {copy_id_button.to_html()}
            </div>
        """  # noqa: E501
        return second_line_html

    def set_and_disable_sync(self) -> None:
        self._sync_checkbox.disabled = True
        self._sync_checkbox.value = True

    def enable_sync(self) -> None:
        if self.show_sync_button:
            self._sync_checkbox.disabled = False

    def set_share_private_data(self) -> None:
        if self.show_share_button:
            self._share_private_checkbox.value = True

    def build(self) -> widgets.VBox:
        content = MainObjectDiffWidget(
            self.diff,
            self.direction,
            with_box=False,
            show_share_warning=self.show_share_button,
            build_state=self.build_state,
        ).widget

        accordion, share_private_checkbox, sync_checkbox = self.build_accordion(
            accordion_body=content,
            show_sync_checkbox=True,
            show_share_private_checkbox=self.show_share_button,
        )

        self._sync_checkbox = sync_checkbox
        self._sync_checkbox.observe(self._on_sync_change, "value")

        self._share_private_checkbox = share_private_checkbox
        self._share_private_checkbox.observe(
            self._on_share_private_data_change, "value"
        )

        return accordion

    def create_accordion_css(
        self, header_id: str, body_id: str, class_name: str
    ) -> str:
        css_accordion = f"""
            <style>
            .accordion {{
                padding: 0 10px;
                margin: 3px 0px;
            }}

            .body-hidden {{
                display: none;
            }}

            .body-visible {{
                display: flex;
            }}

            .{header_id}{{
                display: flex;
                align-items: center;
            }}


            .{class_name}-folded {{
                background: #F4F3F6;
                border: 0.5px solid #B4B0BF;
            }}
            .{class_name}-unfolded {{
                background: white;
                border: 0.5px solid #B4B0BF;
            }}
            </style>
        """
        return css_accordion

    def build_accordion(
        self,
        accordion_body: MainObjectDiffWidget,
        show_sync_checkbox: bool = True,
        show_share_private_checkbox: bool = True,
    ) -> VBox:
        uid = str(uuid4())
        body_id = f"accordion-body-{uid}"
        header_id = f"accordion-header-{uid}"
        class_name = f"accordion-{uid}"
        caret_id = f"caret-{uid}"

        toggle_hide_body_js = f"""
            var body = document.getElementsByClassName('{body_id}')[0];
            var caret = document.getElementById('{caret_id}');
            if (body.classList.contains('body-hidden')) {{
                var vbox = document.getElementsByClassName('{class_name}-folded')[0];
                body.classList.remove('body-hidden');
                body.classList.add('body-visible');
                vbox.classList.remove('{class_name}-folded');
                vbox.classList.add('{class_name}-unfolded');
                caret.classList.remove('fa-caret-right');
                caret.classList.add('fa-caret-down');
            }} else {{
                var vbox = document.getElementsByClassName('{class_name}-unfolded')[0];
                body.classList.remove('body-visible');
                body.classList.add('body-hidden');
                vbox.classList.remove('{class_name}-unfolded');
                vbox.classList.add('{class_name}-folded');
                caret.classList.remove('fa-caret-down');
                caret.classList.add('fa-caret-right');
            }}
        """
        caret = f'<i id="{caret_id}" class="fa fa-fw fa-caret-right"></i>'
        title_html = HTML(
            value=f"<div class='{header_id}' onclick=\"{toggle_hide_body_js}\" style='cursor: pointer; flex-grow: 1; user-select: none; '>{caret} {self.title}</div>",  # noqa: E501
            layout=Layout(flex="1"),
        )

        if isinstance(self.diff.non_empty_object, ActionObject):
            share_data_description = "Share real data and approve"
        else:
            share_data_description = "Share real data"
        share_private_data_checkbox = Checkbox(
            description=share_data_description,
            layout=Layout(width="auto", margin="0 2px 0 0"),
        )
        sync_checkbox = Checkbox(
            description="Sync", layout=Layout(width="auto", margin="0 2px 0 0")
        )

        checkboxes = []
        if show_share_private_checkbox:
            checkboxes.append(share_private_data_checkbox)
        if show_sync_checkbox:
            checkboxes.append(sync_checkbox)

        accordion_header = HBox(
            [title_html] + checkboxes,
            layout=Layout(width="100%", justify_content="space-between"),
        )

        accordion_body.add_class(body_id)
        accordion_body.add_class("body-hidden")

        style = HTML(value=self.create_accordion_css(header_id, body_id, class_name))

        accordion = VBox(
            [style, accordion_header, accordion_body],
            _dom_classes=(f"accordion-{uid}-folded", "accordion"),
        )
        return accordion, share_private_data_checkbox, sync_checkbox

    def _on_sync_change(self, change: Any) -> None:
        self.sync = change["new"]

    def _on_share_private_data_change(self, change: Any) -> None:
        self.share_private_data = change["new"]


class ResolveWidget:
    def __init__(
        self,
        obj_diff_batch: ObjectDiffBatch,
        on_sync_callback: Callable | None = None,
        build_state: bool = True,
    ):
        self.build_state = build_state
        self.obj_diff_batch: ObjectDiffBatch = obj_diff_batch
        self.id2widget: dict[
            UID, CollapsableObjectDiffWidget | MainObjectDiffWidget
        ] = {}
        self.on_sync_callback = on_sync_callback
        self.main_widget = self.build()
        self.result_widget = VBox()  # Placeholder for SyftSuccess / SyftError
        self.widget = VBox(
            [self.build_css_widget(), self.main_widget, self.result_widget]
        )
        self.is_synced = False
        self.hide_result_widget()

    def build_css_widget(self) -> HTML:
        return widgets.HTML(value=CSS_CODE)

    def _repr_mimebundle_(self, **kwargs: dict) -> dict[str, str] | None:
        return self.widget._repr_mimebundle_(**kwargs)

    def click_share_all_private_data(self) -> None:
        for widget in self.id2widget.values():
            widget.set_share_private_data()

    def click_share_private_data(self, uid: UID | str) -> SyftError | SyftSuccess:
        if isinstance(uid, str):
            uid = UID(uid)
        if uid not in self.id2widget:
            return SyftError(message="Object not found in this widget")

        widget = self.id2widget[uid]
        widget.set_share_private_data()
        return SyftSuccess(message="Private data shared")

    def get_share_private_data_state(self) -> dict[UID, bool]:
        return {
            uid: widget.share_private_data for uid, widget in self.id2widget.items()
        }

    def get_mockify_state(self) -> dict[UID, bool]:
        return {uid: widget.mockify for uid, widget in self.id2widget.items()}

    def click_sync(self, *args: list, **kwargs: dict) -> SyftSuccess | SyftError:
        # relative
        from ...client.syncing import handle_sync_batch

        if self.is_synced:
            return SyftError(
                message="The changes in this widget have already been synced."
            )

        res = handle_sync_batch(
            obj_diff_batch=self.obj_diff_batch,
            share_private_data=self.get_share_private_data_state(),
            mockify=self.get_mockify_state(),
        )

        self.set_widget_result_state(res)
        if self.on_sync_callback:
            self.on_sync_callback()
        return res

    @property
    def batch_diff_widgets(self) -> list[CollapsableObjectDiffWidget]:
        dependents = self.obj_diff_batch.get_dependents(
            include_roots=False, include_batch_root=False
        )
        dependent_diff_widgets = [
            CollapsableObjectDiffWidget(
                diff,
                direction=self.obj_diff_batch.sync_direction,
                build_state=self.build_state,
            )
            for diff in dependents
        ]
        return dependent_diff_widgets

    @property
    def dependent_root_diff_widgets(self) -> list[CollapsableObjectDiffWidget]:
        dependencies = self.obj_diff_batch.get_dependencies(
            include_roots=True, include_batch_root=False
        )
        other_roots = [
            d for d in dependencies if d.object_id in self.obj_diff_batch.global_roots
        ]
        widgets = [
            CollapsableObjectDiffWidget(
                diff,
                direction=self.obj_diff_batch.sync_direction,
                build_state=self.build_state,
            )
            for diff in other_roots
        ]
        return widgets

    @property
    def main_object_diff_widget(self) -> MainObjectDiffWidget:
        obj_diff_widget = MainObjectDiffWidget(
            self.obj_diff_batch.root_diff,
            direction=self.obj_diff_batch.sync_direction,
            build_state=self.build_state,
        )
        return obj_diff_widget

    def set_widget_result_state(self, res: SyftSuccess | SyftError) -> None:
        self.is_synced = True
        self.set_result_message(res)
        self.hide_main_widget()
        self.show_result_widget()

    def set_result_message(self, result: SyftSuccess | SyftError) -> None:
        result_html = result._repr_html_()
        # Wrap in div to match Jupyter Lab output styling
        result_html = NOTEBOOK_OUTPUT_DIV.format(content=result_html)
        self.result_widget.children = [widgets.HTML(value=result_html)]

    def hide_main_widget(self) -> None:
        self.main_widget.layout.display = "none"

    def show_main_widget(self) -> None:
        self.main_widget.layout.display = "block"

    def hide_result_widget(self) -> None:
        self.result_widget.layout.display = "none"

    def show_result_widget(self) -> None:
        self.result_widget.layout.display = "block"

    def build(self) -> VBox:
        self.id2widget = {}

        batch_diff_widgets = self.batch_diff_widgets
        dependent_batch_diff_widgets = self.dependent_root_diff_widgets
        main_object_diff_widget = self.main_object_diff_widget

        self.id2widget[main_object_diff_widget.diff.object_id] = main_object_diff_widget

        for widget in batch_diff_widgets:
            self.id2widget[widget.diff.object_id] = widget

        for widget in dependent_batch_diff_widgets:
            self.id2widget[widget.diff.object_id] = widget

        # put a 4px spacer between each item
        main_batch_items = widgets.VBox(
            children=[d.widget for d in batch_diff_widgets],
        )

        dependency_items = widgets.VBox(
            children=[d.widget for d in dependent_batch_diff_widgets],
        )

        full_widget = widgets.VBox(
            [
                self.build_header(),
                self.main_object_diff_widget.widget,
                self.spacer(8),
                main_batch_items,
                self.separator(),
                dependency_items,
                self.spacer(8),
                self.sync_button(),
            ]
        )
        return full_widget

    def sync_button(self) -> Button:
        sync_button = Button(
            description="Apply Selected Changes",
            style={
                "text_color": "#464A91",
                "button_color": "transparent",
                "float": "right",
            },
            layout=Layout(border="#464A91 solid 1.5px", width="200px"),
        )
        sync_button.on_click(self.click_sync)
        return sync_button

    def spacer(self, height: int) -> widgets.HTML:
        return widgets.HTML(f"<div style='height: {height}px'></div>")

    def separator(self) -> widgets.HTML:
        return widgets.HTML(
            value='<div style="text-align: center; margin: 3px 0; border: 1px dashed #B4B0BF;"></div>',
            layout=Layout(width="100%"),
        )

    def build_header(self) -> HTML:
        header_html = SyncWidgetHeader(diff_batch=self.obj_diff_batch).to_html()
        return HTML(value=header_html)


class PaginationControl:
    def __init__(self, data: list, callback: Callable[[int], None]):
        self.data = data
        self.callback = callback
        self.current_index = 0
        self.index_label = widgets.Label(value=f"Index: {self.current_index}")

        self.first_button = widgets.Button(description="First")
        self.previous_button = widgets.Button(description="Previous")
        self.next_button = widgets.Button(description="Next")
        self.last_button = widgets.Button(description="Last")

        self.first_button.on_click(self.go_to_first)
        self.previous_button.on_click(self.go_to_previous)
        self.next_button.on_click(self.go_to_next)
        self.last_button.on_click(self.go_to_last)
        self.output = widgets.Output()

        self.buttons = widgets.HBox(
            [
                self.first_button,
                self.previous_button,
                self.next_button,
                self.last_button,
            ]
        )
        self.update_buttons()
        self.update_index_callback()

    def update_index_label(self) -> None:
        self.index_label.value = f"Current: {self.current_index}"

    def update_buttons(self) -> None:
        self.first_button.disabled = self.current_index == 0
        self.previous_button.disabled = self.current_index == 0
        self.next_button.disabled = self.current_index == len(self.data) - 1
        self.last_button.disabled = self.current_index == len(self.data) - 1

    def go_to_first(self, b: Button | None) -> None:
        self.current_index = 0
        self.update_index_callback()

    def go_to_previous(self, b: Button | None) -> None:
        if self.current_index > 0:
            self.current_index -= 1
        self.update_index_callback()

    def go_to_next(self, b: Button | None) -> None:
        if self.current_index < len(self.data) - 1:
            self.current_index += 1
        self.update_index_callback()

    def go_to_last(self, b: Button | None) -> None:
        self.current_index = len(self.data) - 1
        self.update_index_callback()

    def update_index_callback(self) -> None:
        self.update_index_label()
        self.update_buttons()

        # NOTE self.output is required to display IPython.display.HTML
        # IPython.display.HTML is used to execute JS code
        with self.output:
            self.callback(self.current_index)

    def build(self) -> widgets.VBox:
        return widgets.VBox(
            [widgets.HBox([self.buttons, self.index_label]), self.output]
        )


class PaginatedWidget:
    def __init__(
        self, children: list, on_paginate_callback: Callable[[int], None] | None = None
    ):
        # on_paginate_callback is an optional secondary callback,
        # called after updating the page index and displaying the new widget
        self.children = children
        self.on_paginate_callback = on_paginate_callback
        self.current_index = 0
        self.container = widgets.VBox()

        self.pagination_control = PaginationControl(children, self.on_paginate)

        # Initial display
        self.on_paginate(self.pagination_control.current_index)

    def __getitem__(self, index: int) -> widgets.Widget:
        return self.children[index]

    def on_paginate(self, index: int) -> None:
        self.container.children = [self.children[index]] if self.children else []
        if self.on_paginate_callback:
            self.on_paginate_callback(index)

    def spacer(self, height: int) -> widgets.HTML:
        return widgets.HTML(f"<div style='height: {height}px'></div>")

    def build(self) -> widgets.VBox:
        return widgets.VBox(
            [self.pagination_control.build(), self.spacer(8), self.container]
        )


class PaginatedResolveWidget:
    """
    PaginatedResolveWidget is a widget that displays
    a ResolveWidget for each ObjectDiffBatch,
    paginated by a PaginationControl widget.
    """

    def __init__(self, batches: list[ObjectDiffBatch], build_state: bool = True):
        self.build_state = build_state
        self.batches = batches
        self.resolve_widgets: list[ResolveWidget] = [
            ResolveWidget(
                batch,
                on_sync_callback=partial(self.on_click_sync, i),
                build_state=build_state,
            )
            for i, batch in enumerate(self.batches)
        ]

        self.table_uid = secrets.token_hex(4)

        # Disable the table pagination to avoid the double pagination buttons
        self.batch_table = build_tabulator_table(
            obj=batches,
            uid=self.table_uid,
            max_height=500,
            pagination=False,
            header_sort=False,
        )

        self.paginated_widget = PaginatedWidget(
            children=[widget.widget for widget in self.resolve_widgets],
            on_paginate_callback=self.on_paginate,
        )

        self.table_output = widgets.Output()
        with self.table_output:
            display.display(display.HTML(self.batch_table))
            highlight_single_row(
                self.table_uid, self.paginated_widget.current_index, jump_to_row=True
            )

        self.widget = self.build()

    def on_click_sync(self, index: int) -> None:
        self.update_table_sync_decision(index)
        if self.batches[index].decision is not None:
            self.paginated_widget.pagination_control.go_to_next(None)

    def update_table_sync_decision(self, index: int) -> None:
        new_decision = self.batches[index].decision_badge()
        with self.table_output:
            update_table_cell(
                uid=self.table_uid,
                index=index,
                field="Decision",
                value=new_decision,
            )

    def __getitem__(self, index: int) -> ResolveWidget:
        return self.resolve_widgets[index]

    def on_paginate(self, index: int) -> None:
        return highlight_single_row(self.table_uid, index, jump_to_row=True)

    def build(self) -> widgets.VBox:
        return widgets.VBox([self.table_output, self.paginated_widget.build()])

    def click_sync(self, index: int) -> SyftSuccess | SyftError:
        return self.resolve_widgets[index].click_sync()

    def click_share_all_private_data(self, index: int) -> None:
        self.resolve_widgets[index].click_share_all_private_data()

    def _share_all(self) -> None:
        for widget in self.resolve_widgets:
            widget.click_share_all_private_data()

    def _sync_all(self) -> None:
        for widget in self.resolve_widgets:
            widget.click_sync()

    def _repr_mimebundle_(self, **kwargs: dict) -> dict[str, str] | None:
        return self.widget._repr_mimebundle_(**kwargs)



# stdlib

# third party
from result import Result

# relative
from ...serde.serializable import serializable
from ...server.credentials import SyftVerifyKey
from ...store.document_store import BaseUIDStoreStash
from ...store.document_store import PartitionKey
from ...store.document_store import PartitionSettings
from ...store.document_store import QueryKeys
from ...store.document_store import UIDPartitionKey
from ...types.uid import UID
from ...util.telemetry import instrument
from ..request.request import Request
from ..response import SyftError
from .project import Project

VerifyKeyPartitionKey = PartitionKey(key="user_verify_key", type_=SyftVerifyKey)
NamePartitionKey = PartitionKey(key="name", type_=str)


@instrument
@serializable(canonical_name="ProjectStash", version=1)
class ProjectStash(BaseUIDStoreStash):
    object_type = Project
    settings: PartitionSettings = PartitionSettings(
        name=Project.__canonical_name__, object_type=Project
    )

    def get_all_for_verify_key(
        self, credentials: SyftVerifyKey, verify_key: VerifyKeyPartitionKey
    ) -> Result[list[Request], SyftError]:
        if isinstance(verify_key, str):
            verify_key = SyftVerifyKey.from_string(verify_key)
        qks = QueryKeys(qks=[VerifyKeyPartitionKey.with_obj(verify_key)])
        return self.query_all(
            credentials=credentials,
            qks=qks,
        )

    def get_by_uid(
        self, credentials: SyftVerifyKey, uid: UID
    ) -> Result[Project | None, str]:
        qks = QueryKeys(qks=[UIDPartitionKey.with_obj(uid)])
        return self.query_one(credentials=credentials, qks=qks)

    def get_by_name(
        self, credentials: SyftVerifyKey, project_name: str
    ) -> Result[Project | None, str]:
        qks = QueryKeys(qks=[NamePartitionKey.with_obj(project_name)])
        return self.query_one(credentials=credentials, qks=qks)



# stdlib
from collections.abc import Callable
from collections.abc import Iterator
from datetime import datetime
from datetime import timedelta
import mimetypes
from pathlib import Path
from queue import Queue
import sys
import threading
from time import sleep
from typing import Any
from typing import ClassVar
from typing import TYPE_CHECKING

# third party
from azure.storage.blob import BlobSasPermissions
from azure.storage.blob import generate_blob_sas
from botocore.client import ClientError as BotoClientError
from typing_extensions import Self

# relative
from ..client.api import SyftAPI
from ..client.client import SyftClient
from ..serde import serialize
from ..serde.serializable import serializable
from ..server.credentials import SyftVerifyKey
from ..service.action.action_object import ActionObject
from ..service.action.action_object import ActionObjectPointer
from ..service.action.action_object import BASE_PASSTHROUGH_ATTRS
from ..service.action.action_types import action_types
from ..service.response import SyftError
from ..service.response import SyftException
from ..service.service import from_api_or_context
from ..types.server_url import ServerURL
from ..types.transforms import keep
from ..types.transforms import transform
from .datetime import DateTime
from .syft_object import SYFT_OBJECT_VERSION_1
from .syft_object import SyftObject
from .uid import UID

if TYPE_CHECKING:
    # relative
    from ..store.blob_storage import BlobRetrievalByURL
    from ..store.blob_storage import BlobStorageConnection


READ_EXPIRATION_TIME = 1800  # seconds
DEFAULT_CHUNK_SIZE = 10000 * 1024


@serializable()
class BlobFile(SyftObject):
    __canonical_name__ = "BlobFile"
    __version__ = SYFT_OBJECT_VERSION_1

    file_name: str
    syft_blob_storage_entry_id: UID | None = None
    file_size: int | None = None
    path: Path | None = None
    uploaded: bool = False

    __repr_attrs__ = ["id", "file_name"]

    def read(
        self,
        stream: bool = False,
        chunk_size: int = DEFAULT_CHUNK_SIZE,
        force: bool = False,
    ) -> Any:
        # get blob retrieval object from api + syft_blob_storage_entry_id
        read_method = from_api_or_context(
            "blob_storage.read", self.syft_server_location, self.syft_client_verify_key
        )
        if read_method is not None:
            blob_retrieval_object = read_method(self.syft_blob_storage_entry_id)
            return blob_retrieval_object._read_data(
                stream=stream, chunk_size=chunk_size, _deserialize=False
            )
        else:
            return None

    @classmethod
    def upload_from_path(cls, path: str | Path, client: SyftClient) -> Any:
        # syft absolute
        import syft as sy

        return sy.ActionObject.from_path(path=path).send(client).syft_action_data

    def _upload_to_blobstorage_from_api(self, api: SyftAPI) -> SyftError | None:
        if self.path is None:
            raise ValueError("cannot upload BlobFile, no path specified")
        storage_entry = CreateBlobStorageEntry.from_path(self.path)

        blob_deposit_object = api.services.blob_storage.allocate(storage_entry)

        if isinstance(blob_deposit_object, SyftError):
            return blob_deposit_object

        with open(self.path, "rb") as f:
            result = blob_deposit_object.write(f)

        if isinstance(result, SyftError):
            return result

        self.syft_blob_storage_entry_id = blob_deposit_object.blob_storage_entry_id
        self.uploaded = True

        return None

    def upload_to_blobstorage(self, client: SyftClient) -> SyftError | None:
        self.syft_server_location = client.id
        self.syft_client_verify_key = client.verify_key
        return self._upload_to_blobstorage_from_api(client.api)

    def _iter_lines(self, chunk_size: int = DEFAULT_CHUNK_SIZE) -> Iterator[bytes]:
        """Synchronous version of the async iter_lines. This implementation
        is also optimized in terms of splitting chunks, making it faster for
        larger lines"""
        pending = None
        for chunk in self.read(stream=True, chunk_size=chunk_size):
            if b"\n" in chunk:
                if pending is not None:
                    chunk = pending + chunk  # type: ignore[unreachable]
                lines = chunk.splitlines()
                if lines and lines[-1] and chunk and lines[-1][-1] == chunk[-1]:
                    pending = lines.pop()
                else:
                    pending = None
                yield from lines
            else:
                if pending is None:
                    pending = chunk
                else:
                    pending = pending + chunk

        if pending is not None:
            yield pending

    def read_queue(
        self,
        queue: Queue,
        chunk_size: int,
        progress: bool = False,
        buffer_lines: int = 10000,
    ) -> None:
        total_read = 0
        for _i, line in enumerate(self._iter_lines(chunk_size=chunk_size)):
            line_size = len(line) + 1  # add byte for \n
            if self.file_size is not None:
                total_read = min(self.file_size, total_read + line_size)
            else:
                # naive way of doing this, max be 1 byte off because the last
                # byte can also be a \n
                total_read += line_size
            if progress:
                queue.put((total_read, line))
            else:
                queue.put(line)
            while queue.qsize() > buffer_lines:
                sleep(0.1)
        # Put anything not a string at the end
        queue.put(0)

    def iter_lines(
        self, chunk_size: int = DEFAULT_CHUNK_SIZE, progress: bool = False
    ) -> Iterator[str]:
        item_queue: Queue = Queue()
        threading.Thread(
            target=self.read_queue,
            args=(item_queue, chunk_size, progress),
            daemon=True,
        ).start()
        item = item_queue.get()
        while item != 0:
            yield item
            item = item_queue.get()

    def _coll_repr_(self) -> dict[str, str]:
        return {"file_name": self.file_name}


@serializable(canonical_name="BlobFileType", version=1)
class BlobFileType(type):
    pass


@serializable(canonical_name="BlobFileObjectPointer", version=1)
class BlobFileObjectPointer(ActionObjectPointer):
    pass


@serializable()
class BlobFileObject(ActionObject):
    __canonical_name__ = "BlobFileOBject"
    __version__ = SYFT_OBJECT_VERSION_1

    syft_internal_type: ClassVar[type[Any]] = BlobFile
    syft_pointer_type: ClassVar[type[ActionObjectPointer]] = BlobFileObjectPointer
    syft_passthrough_attrs: list[str] = BASE_PASSTHROUGH_ATTRS


@serializable()
class SecureFilePathLocation(SyftObject):
    __canonical_name__ = "SecureFilePathLocation"
    __version__ = SYFT_OBJECT_VERSION_1

    id: UID
    path: str

    def __repr__(self) -> str:
        return f"{self.path}"

    def generate_url(
        self,
        connection: "BlobStorageConnection",
        type_: type | None,
        bucket_name: str | None,
        *args: Any,
    ) -> "BlobRetrievalByURL":
        raise NotImplementedError


@serializable()
class SeaweedSecureFilePathLocation(SecureFilePathLocation):
    __canonical_name__ = "SeaweedSecureFilePathLocation"
    __version__ = SYFT_OBJECT_VERSION_1

    upload_id: str | None = None

    def generate_url(
        self,
        connection: "BlobStorageConnection",
        type_: type | None,
        bucket_name: str | None,
        *args: Any,
    ) -> "BlobRetrievalByURL":
        try:
            url = connection.client.generate_presigned_url(
                ClientMethod="get_object",
                Params={"Bucket": bucket_name, "Key": self.path},
                ExpiresIn=READ_EXPIRATION_TIME,
            )

            # relative
            from ..store.blob_storage import BlobRetrievalByURL

            return BlobRetrievalByURL(
                url=ServerURL.from_url(url), file_name=Path(self.path).name, type_=type_
            )
        except BotoClientError as e:
            raise SyftException(e)


@serializable()
class AzureSecureFilePathLocation(SecureFilePathLocation):
    __canonical_name__ = "AzureSecureFilePathLocation"
    __version__ = SYFT_OBJECT_VERSION_1

    # upload_id: str
    azure_profile_name: str  # Used by Seaweedfs to refer to a remote config
    bucket_name: str

    def generate_url(
        self, connection: "BlobStorageConnection", type_: type | None, *args: Any
    ) -> "BlobRetrievalByURL":
        # SAS is almost the same thing as the presigned url
        config = connection.config.remote_profiles[self.azure_profile_name]
        account_name = config.account_name
        container_name = config.container_name
        blob_name = self.path
        sas_blob = generate_blob_sas(
            account_name=account_name,
            container_name=container_name,
            blob_name=blob_name,
            account_key=config.account_key,
            permission=BlobSasPermissions(read=True),
            expiry=datetime.utcnow() + timedelta(hours=48),
        )
        url = f"https://{config.account_name}.blob.core.windows.net/{container_name}/{blob_name}?{sas_blob}"

        # relative
        from ..store.blob_storage import BlobRetrievalByURL

        return BlobRetrievalByURL(url=url, file_name=Path(self.path).name, type_=type_)


@serializable()
class BlobStorageEntry(SyftObject):
    __canonical_name__ = "BlobStorageEntry"
    __version__ = SYFT_OBJECT_VERSION_1

    id: UID
    location: SecureFilePathLocation | SeaweedSecureFilePathLocation
    type_: type | None = None
    mimetype: str = "bytes"
    file_size: int
    no_lines: int | None = 0
    uploaded_by: SyftVerifyKey
    created_at: DateTime = DateTime.now()
    bucket_name: str | None = None

    __attr_searchable__ = ["bucket_name"]


@serializable()
class BlobStorageMetadata(SyftObject):
    __canonical_name__ = "BlobStorageMetadata"
    __version__ = SYFT_OBJECT_VERSION_1

    type_: type[SyftObject] | None = None
    mimetype: str = "bytes"
    file_size: int
    no_lines: int | None = 0


@serializable()
class CreateBlobStorageEntry(SyftObject):
    __canonical_name__ = "CreateBlobStorageEntry"
    __version__ = SYFT_OBJECT_VERSION_1

    id: UID
    type_: type | None = None
    mimetype: str = "bytes"
    file_size: int
    extensions: list[str] = []

    @classmethod
    def from_blob_storage_entry(cls, entry: BlobStorageEntry) -> Self:
        # TODO extensions are not stored in the BlobStorageEntry,
        # so a blob entry from path might get a different filename
        # after uploading.
        return cls(
            id=entry.id,
            type_=entry.type_,
            mimetype=entry.mimetype,
            file_size=entry.file_size,
        )

    @classmethod
    def from_obj(cls, obj: SyftObject, file_size: int | None = None) -> Self:
        if file_size is None:
            file_size = sys.getsizeof(serialize._serialize(obj=obj, to_bytes=True))
        return cls(file_size=file_size, type_=type(obj))

    @classmethod
    def from_path(cls, fp: str | Path, mimetype: str | None = None) -> Self:
        path = Path(fp)
        if not path.exists():
            raise SyftException(f"{fp} does not exist.")
        if not path.is_file():
            raise SyftException(f"{fp} is not a file.")

        if path.suffix.lower() == ".jsonl":
            mimetype = "application/json-lines"
        if mimetype is None:
            mime_types = mimetypes.guess_type(fp)
            if len(mime_types) > 0 and mime_types[0] is not None:
                mimetype = mime_types[0]
            else:
                raise SyftException(
                    "mimetype could not be identified.\n"
                    "Please specify mimetype manually `from_path(..., mimetype = ...)`."
                )

        return cls(
            mimetype=mimetype,
            file_size=path.stat().st_size,
            extensions=path.suffixes,
            type_=BlobFileType,
        )

    @property
    def file_name(self) -> str:
        return str(self.id) + "".join(self.extensions)


@transform(BlobStorageEntry, BlobStorageMetadata)
def storage_entry_to_metadata() -> list[Callable]:
    return [keep(["id", "type_", "mimetype", "file_size"])]


action_types[BlobFile] = BlobFileObject



# stdlib
import secrets
from typing import cast

# third party
from kr8s.objects import Service
import yaml

# relative
from ...custom_worker.k8s import KubeUtils
from ...custom_worker.k8s import get_kr8s_client
from ...types.uid import UID
from .rathole import RatholeConfig
from .rathole import get_rathole_port
from .rathole_toml import RatholeClientToml
from .rathole_toml import RatholeServerToml
from .server_peer import ServerPeer

RATHOLE_TOML_CONFIG_MAP = "rathole-config"
RATHOLE_PROXY_CONFIG_MAP = "proxy-config-dynamic"
PROXY_CONFIG_MAP = "proxy-config"
DEFAULT_LOCAL_ADDR_HOST = "0.0.0.0"  # nosec


class RatholeConfigBuilder:
    def __init__(self) -> None:
        self.k8rs_client = get_kr8s_client()

    def add_host_to_server(self, peer: ServerPeer) -> None:
        """Add a host to the rathole server toml file.

        Args:
            peer (ServerPeer): The peer to be added to the rathole server.

        Returns:
            None
        """

        rathole_route = peer.get_rtunnel_route()
        if not rathole_route:
            raise Exception(f"Peer: {peer} has no rathole route: {rathole_route}")

        random_port = self._get_random_port()

        peer_id = cast(UID, peer.id)

        config = RatholeConfig(
            uuid=peer_id.to_string(),
            secret_token=rathole_route.rtunnel_token,
            local_addr_host=DEFAULT_LOCAL_ADDR_HOST,
            local_addr_port=random_port,
            server_name=peer.name,
        )

        # Get rathole toml config map
        rathole_config_map = KubeUtils.get_configmap(
            client=self.k8rs_client, name=RATHOLE_TOML_CONFIG_MAP
        )

        if rathole_config_map is None:
            raise Exception("Rathole config map not found.")

        client_filename = RatholeServerToml.filename

        toml_str = rathole_config_map.data[client_filename]

        # Add the peer info to the toml file
        rathole_toml = RatholeServerToml(toml_str)
        rathole_toml.add_config(config=config)

        # First time adding a peer
        if not rathole_toml.get_rathole_listener_addr():
            bind_addr = f"localhost:{get_rathole_port()}"
            rathole_toml.set_rathole_listener_addr(bind_addr)

        data = {client_filename: rathole_toml.toml_str}

        # Update the rathole config map
        KubeUtils.update_configmap(config_map=rathole_config_map, patch={"data": data})

        # Add the peer info to the proxy config map
        self._add_dynamic_addr_to_rathole(config)

    def remove_host_from_server(self, peer_id: str, server_name: str) -> None:
        """Remove a host from the rathole server toml file.

        Args:
            peer_id (str): The id of the peer to be removed.
            server_name (str): The name of the peer to be removed.

        Returns:
            None
        """

        rathole_config_map = KubeUtils.get_configmap(
            client=self.k8rs_client, name=RATHOLE_TOML_CONFIG_MAP
        )

        if rathole_config_map is None:
            raise Exception("Rathole config map not found.")

        client_filename = RatholeServerToml.filename

        toml_str = rathole_config_map.data[client_filename]

        rathole_toml = RatholeServerToml(toml_str=toml_str)

        rathole_toml.remove_config(peer_id)

        data = {client_filename: rathole_toml.toml_str}

        # Update the rathole config map
        KubeUtils.update_configmap(config_map=rathole_config_map, patch={"data": data})

        # Remove the peer info from the proxy config map
        self._remove_dynamic_addr_from_rathole(server_name)

    def _get_random_port(self) -> int:
        """Get a random port number."""
        return secrets.randbits(15)

    def add_host_to_client(
        self, peer_name: str, peer_id: str, rtunnel_token: str, remote_addr: str
    ) -> None:
        """Add a host to the rathole client toml file."""

        config = RatholeConfig(
            uuid=peer_id,
            secret_token=rtunnel_token,
            local_addr_host="proxy",
            local_addr_port=80,
            server_name=peer_name,
        )

        # Get rathole toml config map
        rathole_config_map = KubeUtils.get_configmap(
            client=self.k8rs_client, name=RATHOLE_TOML_CONFIG_MAP
        )

        if rathole_config_map is None:
            raise Exception("Rathole config map not found.")

        client_filename = RatholeClientToml.filename

        toml_str = rathole_config_map.data[client_filename]

        rathole_toml = RatholeClientToml(toml_str=toml_str)

        rathole_toml.add_config(config=config)

        rathole_toml.set_remote_addr(remote_addr)

        data = {client_filename: rathole_toml.toml_str}

        # Update the rathole config map
        KubeUtils.update_configmap(config_map=rathole_config_map, patch={"data": data})

    def remove_host_from_client(self, peer_id: str) -> None:
        """Remove a host from the rathole client toml file."""

        rathole_config_map = KubeUtils.get_configmap(
            client=self.k8rs_client, name=RATHOLE_TOML_CONFIG_MAP
        )

        if rathole_config_map is None:
            raise Exception("Rathole config map not found.")

        client_filename = RatholeClientToml.filename

        toml_str = rathole_config_map.data[client_filename]

        rathole_toml = RatholeClientToml(toml_str=toml_str)

        rathole_toml.remove_config(peer_id)

        rathole_toml.clear_remote_addr()

        data = {client_filename: rathole_toml.toml_str}

        # Update the rathole config map
        KubeUtils.update_configmap(config_map=rathole_config_map, patch={"data": data})

    def _add_dynamic_addr_to_rathole(
        self, config: RatholeConfig, entrypoint: str = "web"
    ) -> None:
        """Add a port to the rathole proxy config map."""

        rathole_proxy_config_map = KubeUtils.get_configmap(
            self.k8rs_client, RATHOLE_PROXY_CONFIG_MAP
        )

        if rathole_proxy_config_map is None:
            raise Exception("Rathole proxy config map not found.")

        rathole_proxy = rathole_proxy_config_map.data["rathole-dynamic.yml"]

        if not rathole_proxy:
            rathole_proxy = {"http": {"routers": {}, "services": {}, "middlewares": {}}}
        else:
            rathole_proxy = yaml.safe_load(rathole_proxy)

        rathole_proxy["http"]["services"][config.server_name] = {
            "loadBalancer": {
                "servers": [{"url": f"http://rathole:{config.local_addr_port}"}]
            }
        }

        rathole_proxy["http"]["middlewares"]["strip-rathole-prefix"] = {
            "replacePathRegex": {"regex": "^/rathole/(.*)", "replacement": "/$1"}
        }

        proxy_rule = (
            f"Host(`{config.server_name}.syft.local`) || "
            f"HostHeader(`{config.server_name}.syft.local`) && PathPrefix(`/rtunnel`)"
        )

        rathole_proxy["http"]["routers"][config.server_name] = {
            "rule": proxy_rule,
            "service": config.server_name,
            "entryPoints": [entrypoint],
            "middlewares": ["strip-rathole-prefix"],
        }

        KubeUtils.update_configmap(
            config_map=rathole_proxy_config_map,
            patch={"data": {"rathole-dynamic.yml": yaml.safe_dump(rathole_proxy)}},
        )

        self._expose_port_on_rathole_service(config.server_name, config.local_addr_port)

    def _remove_dynamic_addr_from_rathole(self, server_name: str) -> None:
        """Remove a port from the rathole proxy config map."""

        rathole_proxy_config_map = KubeUtils.get_configmap(
            self.k8rs_client, RATHOLE_PROXY_CONFIG_MAP
        )

        if rathole_proxy_config_map is None:
            raise Exception("Rathole proxy config map not found.")

        rathole_proxy = rathole_proxy_config_map.data["rathole-dynamic.yml"]

        if not rathole_proxy:
            return

        rathole_proxy = yaml.safe_load(rathole_proxy)

        if server_name in rathole_proxy["http"]["routers"]:
            del rathole_proxy["http"]["routers"][server_name]

        if server_name in rathole_proxy["http"]["services"]:
            del rathole_proxy["http"]["services"][server_name]

        KubeUtils.update_configmap(
            config_map=rathole_proxy_config_map,
            patch={"data": {"rathole-dynamic.yml": yaml.safe_dump(rathole_proxy)}},
        )

        self._remove_port_on_rathole_service(server_name)

    def _expose_port_on_rathole_service(self, port_name: str, port: int) -> None:
        """Expose a port on the rathole service."""

        rathole_service = KubeUtils.get_service(self.k8rs_client, "rathole")

        rathole_service = cast(Service, rathole_service)

        config = rathole_service.raw

        existing_port_idx = None
        for idx, existing_port in enumerate(config["spec"]["ports"]):
            if existing_port["name"] == port_name:
                print("Port already exists.", existing_port_idx, port_name)
                existing_port_idx = idx
                break

        if existing_port_idx is not None:
            config["spec"]["ports"][existing_port_idx]["port"] = port
            config["spec"]["ports"][existing_port_idx]["targetPort"] = port
        else:
            config["spec"]["ports"].append(
                {
                    "name": port_name,
                    "port": port,
                    "targetPort": port,
                    "protocol": "TCP",
                }
            )

        rathole_service.patch(config)

    def _remove_port_on_rathole_service(self, port_name: str) -> None:
        """Remove a port from the rathole service."""

        rathole_service = KubeUtils.get_service(self.k8rs_client, "rathole")

        rathole_service = cast(Service, rathole_service)

        config = rathole_service.raw

        ports = config["spec"]["ports"]

        for port in ports:
            if port["name"] == port_name:
                ports.remove(port)
                break

        rathole_service.patch(config)



# stdlib
import copy
from typing import Any
from typing import ClassVar

# third party
from typing_extensions import Self

# relative
from ..service.context import AuthedServiceContext
from ..service.response import SyftError
from .syft_object import SYFT_OBJECT_VERSION_1
from .syft_object import SyftObject
from .uid import UID


class SyncableSyftObject(SyftObject):
    __canonical_name__ = "SyncableSyftObject"
    __version__ = SYFT_OBJECT_VERSION_1
    # mapping of private attributes and their mock values
    __private_sync_attr_mocks__: ClassVar[dict[str, any]] = {}

    @classmethod
    def _has_private_sync_attrs(cls: type[Self]) -> bool:
        return len(cls.__private_sync_attr_mocks__) > 0

    def create_shareable_sync_copy(self, mock: bool) -> Self:
        update: dict[str, Any] = {}
        if mock and self._has_private_sync_attrs():
            update |= copy.deepcopy(self.__private_sync_attr_mocks__)
        return self.model_copy(update=update, deep=True)

    def get_sync_dependencies(
        self, context: AuthedServiceContext
    ) -> list[UID] | SyftError:
        return []



# stdlib
from sys import platform

# third party
from capnp.lib.capnp import _DynamicStructBuilder


def get_size(thing: _DynamicStructBuilder | int) -> int:
    if isinstance(thing, int):
        return thing
    return thing.total_size.word_count


def compatible_with_large_file_writes_capnp(thing: _DynamicStructBuilder | int) -> bool:
    if platform in ["darwin", "win32"]:
        return False
    else:
        return get_size(thing) > 50000000  # roughly 0.5GB



# stdlib
from collections.abc import Callable
import inspect
from typing import Any

# relative
from ... import ActionObject
from ... import Worker
from ...client.client import SyftClient
from ...serde.recursive import recursive_serde_register
from ...types.syft_object import SYFT_OBJECT_VERSION_1
from ...types.syft_object import SyftObject
from .action_object import Action
from .action_object import TraceResultRegistry


class Plan(SyftObject):
    __canonical_name__ = "Plan"
    __version__ = SYFT_OBJECT_VERSION_1

    syft_passthrough_attrs: list[str] = [
        "inputs",
        "outputs",
        "code",
        "actions",
        "client",
    ]

    inputs: dict[str, ActionObject]
    outputs: list[ActionObject]
    actions: list[Action]
    code: str
    client: SyftClient | None = None

    def __repr__(self) -> str:
        obj_str = "Plan"

        inp_str = "Inputs:\n"
        inp_str += "\n".join(
            [f"\t\t{k}: {v.__class__.__name__}" for k, v in self.inputs.items()]
        )

        act_str = f"Actions:\n\t\t{len(self.actions)} Actions"

        out_str = "Outputs:\n"
        out_str += "\n".join([f"\t\t{o.__class__.__name__}" for o in self.outputs])

        plan_str = "Plan code:\n"
        plan_str += f'"""\n{self.code}\n"""' if self.code is not None else ""

        return f"{obj_str}\n{inp_str}\n{act_str}\n{out_str}\n\n{plan_str}"

    def remap_actions_to_inputs(self, **new_inputs: Any) -> None:
        pass

    def __call__(self, *args: Any, **kwargs: Any) -> ActionObject | list[ActionObject]:
        if len(self.outputs) == 1:
            return self.outputs[0]
        else:
            return self.outputs


def planify(func: Callable) -> ActionObject:
    TraceResultRegistry.reset_result_for_thread()
    # TraceResult.reset()
    ActionObject.add_trace_hook()
    worker = Worker.named(name="plan_building", reset=True, processes=0)
    client = worker.root_client
    if client is None:
        raise ValueError("Not able to get client for plan building")
    # if client.settings is not None:
    #     client.settings.enable_eager_execution(enable=True) # NOTE: Disabled until we bring back eager execution
    TraceResultRegistry.set_trace_result_for_current_thread(client=client)
    try:
        # TraceResult._client = client
        plan_kwargs = build_plan_inputs(func, client)
        outputs = func(**plan_kwargs)
        if not (isinstance(outputs, list) or isinstance(outputs, tuple)):
            outputs = [outputs]
        ActionObject.remove_trace_hook()
        actions = TraceResultRegistry.get_trace_result_for_thread().result  # type: ignore
        TraceResultRegistry.reset_result_for_thread()
        code = inspect.getsource(func)
        for a in actions:
            if a.create_object is not None:
                # warmup cache
                a.create_object.syft_action_data  # noqa: B018
        plan = Plan(inputs=plan_kwargs, actions=actions, outputs=outputs, code=code)
        return ActionObject.from_obj(plan)
    finally:
        TraceResultRegistry.reset_result_for_thread()


def build_plan_inputs(
    forward_func: Callable, client: SyftClient
) -> dict[str, ActionObject]:
    signature = inspect.signature(forward_func)
    res = {}
    for k, v in signature.parameters.items():
        if v.default is not inspect.Parameter.empty:
            default_value = v.default
            if not isinstance(default_value, ActionObject):
                default_value = ActionObject.from_obj(default_value)
            res[k] = default_value.send(client)
        else:
            raise ValueError(
                f"arg {k} has no placeholder as default value (required for @make_plan functions)"
            )
    return res


recursive_serde_register(Plan)



# stdlib

# stdlib
from collections.abc import Collection
import logging

# relative
from ..abstract_server import ServerSideType
from ..server.credentials import SyftVerifyKey
from ..service.response import SyftError
from ..service.response import SyftSuccess
from ..service.sync.diff_state import ObjectDiffBatch
from ..service.sync.diff_state import ServerDiff
from ..service.sync.diff_state import SyncInstruction
from ..service.sync.resolve_widget import PaginatedResolveWidget
from ..service.sync.resolve_widget import ResolveWidget
from ..service.sync.sync_state import SyncState
from ..types.uid import UID
from ..util.decorators import deprecated
from ..util.util import prompt_warning_message
from .datasite_client import DatasiteClient
from .sync_decision import SyncDecision
from .sync_decision import SyncDirection

logger = logging.getLogger(__name__)


def sync(
    from_client: DatasiteClient,
    to_client: DatasiteClient,
    include_ignored: bool = False,
    include_same: bool = False,
    filter_by_email: str | None = None,
    include_types: Collection[str | type] | None = None,
    exclude_types: Collection[str | type] | None = None,
    hide_usercode: bool = True,
) -> PaginatedResolveWidget | SyftError | SyftSuccess:
    diff = compare_clients(
        from_client=from_client,
        to_client=to_client,
        include_ignored=include_ignored,
        include_same=include_same,
        filter_by_email=filter_by_email,
        include_types=include_types,
        exclude_types=exclude_types,
        hide_usercode=hide_usercode,
    )
    if isinstance(diff, SyftError):
        return diff

    return diff.resolve()


def compare_states(
    from_state: SyncState,
    to_state: SyncState,
    include_ignored: bool = False,
    include_same: bool = False,
    filter_by_email: str | None = None,
    include_types: Collection[str | type] | None = None,
    exclude_types: Collection[str | type] | None = None,
    hide_usercode: bool = True,
) -> ServerDiff | SyftError:
    # ServerDiff
    if (
        from_state.server_side_type == ServerSideType.LOW_SIDE
        and to_state.server_side_type == ServerSideType.HIGH_SIDE
    ):
        low_state = from_state
        high_state = to_state
        direction = SyncDirection.LOW_TO_HIGH
    elif (
        from_state.server_side_type == ServerSideType.HIGH_SIDE
        and to_state.server_side_type == ServerSideType.LOW_SIDE
    ):
        low_state = to_state
        high_state = from_state
        direction = SyncDirection.HIGH_TO_LOW
    else:
        return SyftError(
            "Invalid server side types: can only compare a high and low server"
        )

    if hide_usercode:
        prompt_warning_message(
            "UserCodes are hidden by default, and are part of the Requests."
            " If you want to include them as separate objects, set `hide_usercode=False`"
        )
        exclude_types = exclude_types or []
        exclude_types.append("usercode")

    return ServerDiff.from_sync_state(
        low_state=low_state,
        high_state=high_state,
        direction=direction,
        include_ignored=include_ignored,
        include_same=include_same,
        filter_by_email=filter_by_email,
        include_types=include_types,
        exclude_types=exclude_types,
    )


def compare_clients(
    from_client: DatasiteClient,
    to_client: DatasiteClient,
    include_ignored: bool = False,
    include_same: bool = False,
    filter_by_email: str | None = None,
    include_types: Collection[str | type] | None = None,
    exclude_types: Collection[str | type] | None = None,
    hide_usercode: bool = True,
) -> ServerDiff | SyftError:
    from_state = from_client.get_sync_state()
    if isinstance(from_state, SyftError):
        return from_state

    to_state = to_client.get_sync_state()
    if isinstance(to_state, SyftError):
        return to_state

    return compare_states(
        from_state=from_state,
        to_state=to_state,
        include_ignored=include_ignored,
        include_same=include_same,
        filter_by_email=filter_by_email,
        include_types=include_types,
        exclude_types=exclude_types,
        hide_usercode=hide_usercode,
    )


def resolve(
    obj: ObjectDiffBatch | ServerDiff,
) -> ResolveWidget | PaginatedResolveWidget | SyftSuccess | SyftError:
    if not isinstance(obj, ObjectDiffBatch | ServerDiff):
        raise ValueError(
            f"Invalid type: could not resolve object with type {type(obj).__qualname__}"
        )
    return obj.resolve()


@deprecated(reason="resolve_single has been renamed to resolve", return_syfterror=True)
def resolve_single(
    obj_diff_batch: ObjectDiffBatch,
) -> ResolveWidget | PaginatedResolveWidget | SyftSuccess | SyftError:
    return resolve(obj_diff_batch)


def handle_sync_batch(
    obj_diff_batch: ObjectDiffBatch,
    share_private_data: dict[UID, bool],
    mockify: dict[UID, bool],
) -> SyftSuccess | SyftError:
    # Infer SyncDecision
    sync_direction = obj_diff_batch.sync_direction
    if sync_direction is None:
        return SyftError(
            message="Cannot sync an object without a specified sync direction."
        )

    decision = sync_direction.to_sync_decision()

    # Validate decision
    if decision not in [SyncDecision.LOW, SyncDecision.HIGH]:
        return SyftError(message="Invalid sync decision")
    elif obj_diff_batch.is_unchanged:
        return SyftSuccess(message="No changes to sync")
    elif obj_diff_batch.decision is SyncDecision.IGNORE:
        return SyftError(
            message="Attempted to sync an ignored object, please unignore first"
        )
    elif obj_diff_batch.decision is not None:
        return SyftError(
            message="Attempted to sync an object that has already been synced"
        )

    src_client = obj_diff_batch.source_client
    tgt_client = obj_diff_batch.target_client
    src_resolved_state, tgt_resolved_state = obj_diff_batch.create_new_resolved_states()

    obj_diff_batch.decision = decision

    sync_instructions = []
    for diff in obj_diff_batch.get_dependencies(include_roots=True):
        # figure out the right verify key to share to
        # in case of a job with user code, share to user code owner
        # without user code, share to job owner
        share_to_user: SyftVerifyKey | None = (
            getattr(obj_diff_batch.user_code_high, "user_verify_key", None)
            or obj_diff_batch.user_verify_key_high
        )
        share_private_data_for_diff = share_private_data[diff.object_id]
        mockify_for_diff = mockify[diff.object_id]
        instruction = SyncInstruction.from_batch_decision(
            diff=diff,
            share_private_data=share_private_data_for_diff,
            mockify=mockify_for_diff,
            sync_direction=sync_direction,
            decision=decision,
            share_to_user=share_to_user,
        )
        sync_instructions.append(instruction)

    logger.debug(f"Decision: Syncing {len(sync_instructions)} objects")

    # Apply empty state to source side to signal that we are done syncing
    res_src = src_client.apply_state(src_resolved_state)
    if isinstance(res_src, SyftError):
        return res_src

    # Apply sync instructions to target side
    for sync_instruction in sync_instructions:
        tgt_resolved_state.add_sync_instruction(sync_instruction)
    res_tgt = tgt_client.apply_state(tgt_resolved_state)

    return res_tgt


def handle_ignore_batch(
    obj_diff_batch: ObjectDiffBatch,
    all_batches: list[ObjectDiffBatch],
) -> SyftSuccess | SyftError:
    if obj_diff_batch.decision is SyncDecision.IGNORE:
        return SyftSuccess(message="This batch is already ignored")
    elif obj_diff_batch.decision is not None:
        return SyftError(
            message="Attempted to sync an object that has already been synced"
        )

    obj_diff_batch.decision = SyncDecision.IGNORE
    other_batches = [b for b in all_batches if b is not obj_diff_batch]
    other_ignore_batches = get_other_ignore_batches(obj_diff_batch, other_batches)

    for other_batch in other_ignore_batches:
        other_batch.decision = SyncDecision.IGNORE
        logger.debug(f"Ignoring other batch with root {other_batch.root_type.__name__}")

    src_client = obj_diff_batch.source_client
    tgt_client = obj_diff_batch.target_client
    src_resolved_state, tgt_resolved_state = obj_diff_batch.create_new_resolved_states()

    for batch in [obj_diff_batch] + other_ignore_batches:
        src_resolved_state.add_ignored(batch)
        tgt_resolved_state.add_ignored(batch)

    res_src = src_client.apply_state(src_resolved_state)
    if isinstance(res_src, SyftError):
        return res_src

    res_tgt = tgt_client.apply_state(tgt_resolved_state)
    return res_tgt


def handle_unignore_batch(
    obj_diff_batch: ObjectDiffBatch,
    all_batches: list[ObjectDiffBatch],
) -> SyftSuccess | SyftError:
    src_client = obj_diff_batch.source_client
    tgt_client = obj_diff_batch.target_client
    src_resolved_state, tgt_resolved_state = obj_diff_batch.create_new_resolved_states()

    obj_diff_batch.decision = None
    src_resolved_state.add_unignored(obj_diff_batch.root_id)
    tgt_resolved_state.add_unignored(obj_diff_batch.root_id)

    # Unignore dependencies
    other_batches = [b for b in all_batches if b is not obj_diff_batch]
    other_unignore_batches = get_other_unignore_batches(obj_diff_batch, other_batches)
    for other_batch in other_unignore_batches:
        logger.debug(f"Ignoring other batch with root {other_batch.root_type.__name__}")
        other_batch.decision = None
        src_resolved_state.add_unignored(other_batch.root_id)
        tgt_resolved_state.add_unignored(other_batch.root_id)

    res_src = src_client.apply_state(src_resolved_state)
    if isinstance(res_src, SyftError):
        return res_src

    res_tgt = tgt_client.apply_state(tgt_resolved_state)
    return res_tgt


def get_other_unignore_batches(
    batch: ObjectDiffBatch,
    other_batches: list[ObjectDiffBatch],
) -> list[ObjectDiffBatch]:
    if batch.decision is not None:
        return []

    other_unignore_batches = []
    required_dependencies = {
        d.object_id for d in batch.get_dependencies(include_roots=True)
    }

    for other_batch in other_batches:
        if other_batch == batch:
            continue
        elif (
            other_batch.decision == SyncDecision.IGNORE
            and other_batch.root_id in required_dependencies
        ):
            other_unignore_batches.append(other_batch)
    return other_unignore_batches


def get_other_ignore_batches(
    batch: ObjectDiffBatch,
    other_batches: list[ObjectDiffBatch],
) -> list[ObjectDiffBatch]:
    if batch.decision != SyncDecision.IGNORE:
        return []

    other_ignore_batches = []
    ignored_ids = {x.object_id for x in batch.get_dependents(include_roots=False)}
    for other_batch in other_batches:
        if other_batch.decision != SyncDecision.IGNORE:
            # Currently, this is not recursive, in the future it might be
            other_batch_ids = {
                d.object_id for d in other_batch.get_dependencies(include_roots=True)
            }
            if len(other_batch_ids & ignored_ids) != 0:
                other_ignore_batches.append(other_batch)
                ignored_ids.update(other_batch_ids)

    return other_ignore_batches



# stdlib
from typing import Any

# relative
from ...serde.serializable import serializable
from ...store.document_store import SYFT_OBJECT_VERSION_1
from ...store.document_store import SyftObject
from ...types.datetime import DateTime


@serializable()
class DockerWorker(SyftObject):
    # version
    __canonical_name__ = "ContainerImage"
    __version__ = SYFT_OBJECT_VERSION_1

    __attr_searchable__ = ["container_id", "container_name"]
    __attr_unique__ = ["container_id"]
    __repr_attrs__ = ["container_id", "created_at"]

    container_name: str
    container_id: str
    created_at: DateTime = DateTime.now()

    def _coll_repr_(self) -> dict[str, Any]:
        return {
            "container_name": self.container_name,
            "container_id": self.container_id,
            "created_at": self.created_at,
        }
